from typing import List, Optional

import numpy as np
import tensornetwork as tn
import networkx as nx
import matplotlib.pyplot as plt
from ttnopt.src.functionTTN import (
    get_renormalization_sequence,
)
from copy import copy


class TreeTensorNetwork:
    """A class for Tree Tensor Network (TTN)."""

    def __init__(
        self,
        edges: List[List[int]],
        tensors: Optional[List[np.ndarray]] = None,
        top_edge_id: Optional[int] = None,
        gauge_tensor: Optional[np.ndarray] = None,
        norm: Optional[float] = None,
    ):
        """Initialize a TreeTensorNetwork object.

        Args:
            edges (List[List[int]]): Edge id list for each tensor in the order [left, right, top].
            top_edge_id (int): edge id that connects to the top tensor.
            tensors (Optional[List[np.ndarray]]): tensors for each node.
            gauge_tensor (Optional[np.ndarray]): gauge tensor at top_edge_bond.
                This parameter is not required for some algorithms (Ground State Search, etc.)
        """
        self.edges = edges

        self.tensor_num = len(edges)
        self.physical_edges = self._physical_edges()
        self.size = len(self.physical_edges)

        self.top_edge_id = None
        self.canonical_center_edge_id = None
        self.tensors = None
        self.gauge_tensor = None
        self.norm = 1.0
        self.edge_dims = {}
        if top_edge_id is not None:
            self.top_edge_id = top_edge_id
            self.canonical_center_edge_id = top_edge_id
        else:
            dict_edge = {edge[2]: 0 for edge in edges}
            for e in edges:
                dict_edge[e[2]] += 1
            self.top_edge_id = [k for k, v in dict_edge.items() if v == 2][0]
            self.canonical_center_edge_id = copy(self.top_edge_id)
        if tensors is not None:
            self.tensors = tensors
            self.edge_dims = self._edge_dims()
        if gauge_tensor is not None:
            self.gauge_tensor = gauge_tensor
        if norm is not None:
            self.norm = norm

    @classmethod
    def mps(
        cls,
        size: int,
        target: Optional[np.ndarray] = None,
        max_bond_dimension: Optional[int] = None,
    ):
        """Initialize an State object with matrix product structure.
        Args:
            size : The size of system.
        """

        edges = []
        upper_edge_id = size
        edges.append([0, 1, upper_edge_id])
        for i in range(2, (size - 2) // 2 + 1):
            edges.append([upper_edge_id, i, upper_edge_id + 1])
            upper_edge_id += 1

        center_edge_id = upper_edge_id
        upper_edge_id = upper_edge_id + 1

        tmp_edges = []
        tmp_edges.append([size - 2, size - 1, upper_edge_id])

        for i in reversed(range((size - 2) // 2 + 2, size - 2)):
            tmp_edges.append([i, upper_edge_id, upper_edge_id + 1])
            upper_edge_id += 1

        tmp_edges.append([(size - 2) // 2 + 1, upper_edge_id, center_edge_id])
        edges += reversed(tmp_edges)

        if target is not None and max_bond_dimension is not None:
            norm = np.linalg.norm(target)
            normed_target = target / norm

            if len(target.shape) != size:
                raise ValueError(
                    f"The shape of the tensor is not correct. tensor.shape={target.shape}, size={size}"
                )
            target_node = tn.Node(normed_target)
            U, S, V, _ = tn.split_node_full_svd(
                target_node,
                target_node[:2],
                target_node[2:],
                max_singular_values=max_bond_dimension,
            )
            tensors = [U.tensor]
            V = tn.contractors.auto([V, S], [S[0]] + V[1:])
            for i in range(2, (size - 2) // 2 + 1):
                U, S, V, _ = tn.split_node_full_svd(
                    V, [V[0], V[1]], V[2:], max_singular_values=max_bond_dimension
                )
                tensors.append(U.tensor)
                V = tn.contractors.auto([V, S], [S[0]] + V[1:])

            U, S, V, _ = tn.split_node_full_svd(
                V, V[:-2], [V[-2], V[-1]], max_singular_values=max_bond_dimension
            )
            tmp_tensors = [V.reorder_edges([V[1], V[2], V[0]]).tensor]
            U = tn.contractors.auto([U, S], U[:-1] + [S[1]])
            for i in reversed(range((size - 2) // 2 + 2, size - 2)):
                U, S, V, _ = tn.split_node_full_svd(
                    U, U[:-2], [U[-2], U[-1]], max_singular_values=max_bond_dimension
                )
                U = tn.contractors.auto([U, S], U[:-1] + [S[1]])
                tmp_tensors.append(V.reorder_edges([V[1], V[2], V[0]]).tensor)
            U, S, V, _ = tn.split_node_full_svd(
                U, U[:-2], [U[-2], U[-1]], max_singular_values=max_bond_dimension
            )
            tmp_tensors.append(V.reorder_edges([V[1], V[2], V[0]]).tensor)
            tensors += reversed(tmp_tensors)
            return cls(edges, tensors, center_edge_id, S.tensor, norm.item())

        return cls(edges, top_edge_id=center_edge_id)

    @classmethod
    def tree(cls, size: int):
        """Initialize an object with binary tree structure.

        Args:
            size : The size of system.
        """
        edges = []

        layer_index = 0
        num_layer = int(np.log2(size)) - 1
        for layer in range(num_layer):
            tensor_num = int(2 ** (np.log2(size) - 1 - layer))
            nn = int(2 ** (np.log2(size) - layer))
            for i in range(tensor_num):
                if layer != num_layer - 1:
                    edge_id = [
                        layer_index + i * 2,
                        layer_index + i * 2 + 1,
                        layer_index + nn + i,
                    ]

                else:
                    edge_id = [
                        layer_index + i * 2,
                        layer_index + i * 2 + 1,
                        layer_index + nn,
                    ]

                edges.append(edge_id)

            layer_index += nn

        center_edge_id = layer_index

        return cls(edges, top_edge_id=center_edge_id)

    @classmethod
    def init_random(
        self,
        edges: List[List[int]],
        top_edge_id: Optional[int] = None,
        edge_dims: Optional[dict] = None,
        init_bond_dimension: int = 4,
    ):
        self.tensors = []
        for _ in edges:
            self.tensors.append(np.array([]))
        sequence = get_renormalization_sequence(edges, top_edge_id)
        for tensor_id in sequence:
            m = (
                self.edge_dims[edges[tensor_id][0]]
                * self.edge_dims[edges[tensor_id][1]]
            )
            n = np.min([m, init_bond_dimension])
            random_matrix = np.random.normal(0, 1, (m, n))
            Q, _ = np.linalg.qr(random_matrix)
            self.tensors[tensor_id] = np.reshape(
                Q,
                (
                    self.edge_dims[self.edges[tensor_id][0]],
                    self.edge_dims[self.edges[tensor_id][1]],
                    n,
                ),
            )
            self.edge_dims[self.edges[tensor_id][2]] = n
        self.gauge_tensor = np.eye(n)

    def visualize(self):
        """Visualize the TreeTensorNetwork."""
        g = nx.DiGraph()
        logs = self._get_parent_child_pairs()

        small_black_nodes = []
        large_red_nodes = []
        default_nodes = []
        for log in logs:  # log=[self node, parent node]
            g.add_node(log[0])
            g.add_edge(log[1], log[0])
            # ノードの種類に応じて分類
            if "bare" in log[0]:
                small_black_nodes.append(log[0])
            else:
                default_nodes.append(log[0])
            if log[1] == "top":
                large_red_nodes.append(log[1])
        pos = nx.nx_pydot.graphviz_layout(g, prog="twopi")
        # matplotlib settings
        fig, ax = plt.subplots(figsize=(5, 5), dpi=300)

        # isometry node
        nx.draw(
            g,
            ax=ax,
            pos=pos,
            nodelist=default_nodes,
            with_labels=False,
            arrows=False,
            node_size=20,
            node_shape="o",
            width=0.5,
            node_color="blue",
        )

        # bare node
        nx.draw_networkx_nodes(
            g,
            ax=ax,
            pos=pos,
            nodelist=small_black_nodes,
            node_size=8,
            node_shape="o",
            node_color="black",
        )

        # top node
        # nx.draw_networkx_nodes(
        #     g,
        #     ax=ax,
        #     pos=pos,
        #     nodelist=large_red_nodes,
        #     node_size=25,
        #     node_shape="o",
        #     node_color="red",
        # )
        return plt

    def central_tensor_ids(self):
        central_tensor_ids = [
            tensor_id
            for tensor_id, edges in enumerate(self.edges)
            if edges[2] == self.canonical_center_edge_id
        ]
        return central_tensor_ids

    def _physical_edges(self):
        count_dict = {}
        for edge in self.edges:
            for i in edge:
                if i not in count_dict.keys():
                    count_dict[i] = 0
                count_dict[i] += 1
        physical_edges = [k for k, v in count_dict.items() if v == 1]
        return physical_edges

    def _edge_dims(self):
        edge_dims = {}
        for i, t in enumerate(self.tensors):
            for j, e in enumerate(self.edges[i]):

                edge_dims[e] = t.shape[j]
        return edge_dims

    def _get_parent_child_pairs(self):
        parent_child_pairs = []
        structure_edges = self.edges

        for i, edges in enumerate(structure_edges):
            child1_edge, child2_edge, parent_edge = edges
            if child1_edge in self.physical_edges:
                parent_child_pairs.append([f"bare{child1_edge}", str(i)])
            if child2_edge in self.physical_edges:
                parent_child_pairs.append([f"bare{child2_edge}", str(i)])
            for j, _edges in enumerate(structure_edges):
                if j != i and parent_edge in _edges[:2]:
                    parent_child_pairs.append([str(i), str(j)])
                if parent_edge == self.canonical_center_edge_id:
                    parent_child_pairs.append([str(i), "top"])
        return parent_child_pairs



from typing import Dict, Union

import numpy as np
import tensornetwork as tn
import copy

from ttnopt.src.DataEngine import DataEngine
from ttnopt.src.TTN import TreeTensorNetwork


class FactorizeTensor(DataEngine):
    """A class for ground state search algorithm based on DMRG.
    Args:
        psi: The instance of TTN Class
        target (np.array, optional): The target tensor
        init_bond_dim (int, optional): The bond dimension which are used to initialize tensors
        max_bond_dim (int, optional): The maximum bond dimension during updating tensors
        truncation_err (float, optional): The maximum truncation error during updating tensors
    """

    def __init__(
        self,
        psi: TreeTensorNetwork,
        max_bond_dim: int = 16,
    ):
        """Initialize a FactorizeTensor object.

        Args:
            psi (TreeTensorNetwork): The quantum state.
            target (np.ndarray): Target tensor.
            init_bond_dim (int, optional): Initial bond dimension. Defaults to 4.
            max_bond_dim (int, optional): Maximum bond dimension. Defaults to 16.
        """
        self.entanglement: Dict[int, float] = {}
        self.fidelity: Dict[int, float] = {}
        self.error: Dict[int, float] = {}
        self.bond_dim: Dict[int, int] = {}
        super().__init__(psi, max_bond_dim)

    def run(
        self,
        target: Union[np.ndarray, None] = None,
        opt_fidelity=False,
        opt_structure=0,
        fidelity_convergence_threshold=1e-8,
        entanglement_convergence_threshold=1e-8,
        max_num_sweep=5,
        converged_count=2,
        max_truncation_error=0.0,
        temperature: float = 0.0,
        tau: int = 0,
    ):
        """Run FactorizingTensor algorithm.

        Args:
            target (np.ndarray): Target tensor.
            opt_fidelity (bool, optional): If optimize the fidelity or not. Defaults to False.
            opt_structure (bool, optional): If optimize the tree structure or not. Defaults to False.
            fidelity_convergence_threshold (float, optional): Fidelity threshold for convergence. Defaults to 1e-8.
            entanglement_convergence_threshold (float, optional): Entanglement entropy threshold for automatic optimization. Defaults to 1e-8.
            max_num_sweep (int, optional): Maximum number of sweeps. Defaults to 5.
            converged_count (int, optional): Converged count. Defaults to 2.
            max_truncation_error (float, optimal): .
        """

        _ee_at_edge: Dict[int, float] = {}
        ee_at_edge: Dict[int, float] = {}
        _fidelity_at_edge: Dict[int, float] = {}
        fidelity_at_edge: Dict[int, float] = {}
        _error_at_edge: Dict[int, float] = {}
        bond_dim: Dict[int, int] = {}

        edges, _edges = copy.deepcopy(self.psi.edges), copy.deepcopy(self.psi.edges)

        converged_num = 0

        for sweep_num in range(max_num_sweep):
            temp = temperature * (2 ** (-sweep_num / tau))
            if converged_num > converged_count:
                break

            ee_at_edge = copy.deepcopy(_ee_at_edge)
            fidelity_at_edge = copy.deepcopy(_fidelity_at_edge)
            edges = copy.deepcopy(_edges)

            self.distance = self.initial_distance()
            self.flag = self.initial_flag()

            print("Sweep count: " + str(sweep_num))
            while self.candidate_edge_ids() != []:
                (
                    edge_id,
                    selected_tensor_id,
                    connected_tensor_id,
                    not_selected_tensor_id,
                ) = self.local_two_tensor()

                self.set_flag(not_selected_tensor_id)

                if opt_fidelity:
                    self.set_ttn_properties_at_one_tensor(edge_id, selected_tensor_id)
                    new_tensor = self.update_tensor(
                        target, [selected_tensor_id, connected_tensor_id]
                    )
                else:
                    iso1 = tn.Node(self.psi.tensors[selected_tensor_id])
                    gauge = tn.Node(self.psi.gauge_tensor)
                    iso1[2] ^ gauge[0]
                    iso1 = tn.contractors.auto(
                        [iso1, gauge], output_edge_order=[iso1[0], iso1[1], gauge[1]]
                    )
                    self.psi.tensors[selected_tensor_id] = iso1.tensor
                    self.set_ttn_properties_at_one_tensor(edge_id, selected_tensor_id)

                    iso1 = tn.Node(self.psi.tensors[selected_tensor_id])
                    iso2 = tn.Node(self.psi.tensors[connected_tensor_id])
                    iso1[2] ^ iso2[2]
                    new_tensor = tn.contractors.auto(
                        [iso1, iso2],
                        output_edge_order=[iso1[0], iso1[1], iso2[0], iso2[1]],
                    )

                psi_edges = (
                    self.psi.edges[selected_tensor_id][:2]
                    + self.psi.edges[connected_tensor_id][:2]
                )
                u, s, v, probability, error, edge_order = self.decompose_two_tensors(
                    new_tensor,
                    self.max_bond_dim,
                    opt_structure=opt_structure,
                    temperature=temp,
                    operate_degeneracy=False,
                    max_truncation_error=max_truncation_error,
                )

                self.psi.tensors[selected_tensor_id] = u
                self.psi.tensors[connected_tensor_id] = v
                self.psi.gauge_tensor = s
                (
                    self.psi.edges[selected_tensor_id][0],
                    self.psi.edges[selected_tensor_id][1],
                ) = (
                    psi_edges[edge_order[0]],
                    psi_edges[edge_order[1]],
                )
                (
                    self.psi.edges[connected_tensor_id][0],
                    self.psi.edges[connected_tensor_id][1],
                ) = (
                    psi_edges[edge_order[2]],
                    psi_edges[edge_order[3]],
                )

                self.distance = self.initial_distance()
                if target is not None:
                    fidelity = self.get_fidelity(
                        target,
                    )
                    print("Fidelity:", fidelity)
                    _fidelity_at_edge[self.psi.canonical_center_edge_id] = fidelity
                ee = self.entanglement_entropy(probability)
                _ee_at_edge[self.psi.canonical_center_edge_id] = ee
                ee_dict = self.entanglement_entropy_at_physical_bond(
                    new_tensor, psi_edges
                )
                for key in ee_dict.keys():
                    _ee_at_edge[key] = ee_dict[key]
                _error_at_edge[self.psi.canonical_center_edge_id] = error

                bond_dim[self.psi.canonical_center_edge_id] = (
                    self.psi.gauge_tensor.shape[0]
                )

            _edges = copy.deepcopy(self.psi.edges)

            # 終了判定
            sweep_num += 1
            if sweep_num > 2:
                diff_ee = [
                    np.abs(ee_at_edge[key] - _ee_at_edge[key])
                    for key in ee_at_edge.keys()
                ]
                diff_fidelity = [
                    np.abs(fidelity_at_edge[key] - _fidelity_at_edge[key])
                    for key in fidelity_at_edge.keys()
                ]
                if all(
                    [
                        set(edge[:2]) == set(_edge[:2]) and edge[2] == _edge[2]
                        for edge, _edge in zip(edges, _edges)
                    ]
                ):
                    if all(
                        [ee < entanglement_convergence_threshold for ee in diff_ee]
                    ) and all(
                        [
                            diff_fidelity < fidelity_convergence_threshold
                            for diff_fidelity in diff_fidelity
                        ]
                    ):
                        converged_num += 1
        print("Converged")

        self.entanglement = _ee_at_edge
        self.fidelity = _fidelity_at_edge
        self.error = _error_at_edge
        self.bond_dim = bond_dim

        return 0



from copy import deepcopy
from typing import Dict, Tuple, Union

import numpy as np
import tensornetwork as tn
from tensornetwork import BlockSparseTensor, Index, U1Charge

from ttnopt.src.Hamiltonian import Hamiltonian
from ttnopt.src.PhysicsEngineSparse import PhysicsEngineSparse
from ttnopt.src.TTN import TreeTensorNetwork

# tn.block_sparse.enable_caching()


class GroundStateSearchSparse(PhysicsEngineSparse):
    """A class for ground state search algorithm based on DMRG using Sparse Tensor.
    Args:
        psi (TreeTensorNetwork): The quantum state.
        hamiltonians (Hamiltonian): Hamiltonian which is list of Observable.
        u1_num (str): The number of preserving total spin in U(1) symmetry.
        init_bond_dim (int, optional): Initial bond dimension. Defaults to 4.
        max_bond_dim (int, optional): Maximum bond dimension. Defaults to 16.
        truncation_error (float, optional): Maximum truncation error. Defaults to 1e-11.
    """

    def __init__(
        self,
        psi: TreeTensorNetwork,
        hamiltonian: Hamiltonian,
        u1_num: Union[int, str],
        init_bond_dim: int = 4,
        max_bond_dim: int = 16,
        energy_degeneracy_threshold: float = 1e-13,
        entanglement_degeneracy_threshold: float = 0.1,
    ):
        """Initialize a DMRG object.

        Args:
            psi : The quantum state.
            hamiltonians : The Hamiltonian.
            u1_num : The number of preserving total spin in U(1) symmetry.
            init_bond_dim : Initial bond dimension.
            max_bond_dim : Maximum bond dimension.
            truncation_error : Maximum truncation error.
        """
        self.energy: Dict[int, float] = {}
        self.entanglement: Dict[int, float] = {}
        self.error: Dict[int, float] = {}
        self.one_site_expval: Dict[int, Dict[str, float]] = {}
        self.two_site_expval: Dict[Tuple[int, int], Dict[str, float]] = {}

        # set backend only in this function
        super().__init__(
            psi,
            hamiltonian,
            u1_num,
            init_bond_dim,
            max_bond_dim,
            energy_degeneracy_threshold,
            entanglement_degeneracy_threshold,
        )

    def run(
        self,
        opt_structure: int = 0,
        energy_convergence_threshold: float = 1e-10,
        entanglement_convergence_threshold: float = 1e-10,
        max_num_sweep: int = 10,
        converged_count: int = 2,
        eval_onesite_expval: bool = False,
        eval_twosite_expval: bool = False,
        sz_sign: int = 0,
        temperature: float = 0.0,
        tau: int = 1,
    ):
        """Run Ground State Search algorithm using Sparse Tensor with total spin.

        Args:
            opt_structure (bool, optional): If optimize the tree structure or not. Defaults to False.
            energy_convergence_threshold (float, optional): Energy threshold for convergence. Defaults to 1e-8.
            entanglement_convergence_threshold (float, optional): Entanglement entropy threshold for automatic optimization. Defaults to 1e-8.
            converged_count (int, optional): Converged count. Defaults to 1.
            eval_onesite_expval (bool): If evaluate one-site expectation value or not.
            eval_twosite_expval (bool): If evaluate two-site expectation value or not.
        """
        energy_at_edge: Dict[int, float] = {}
        _energy_at_edge: Dict[int, float] = {}
        ee_at_edge: Dict[int, float] = {}
        _ee_at_edge: Dict[int, float] = {}
        _error_at_edge: Dict[int, float] = {}
        onesite_expval: Dict[int, Dict[str, float]] = {}
        twosite_expval: Dict[Tuple[int, int], Dict[str, float]] = {}

        edges, _edges = deepcopy(self.psi.edges), deepcopy(self.psi.edges)

        converged_num = 0

        if tau == 0:
            tau = max_num_sweep // 2
        for sweep_num in range(max_num_sweep):
            temp = temperature * (2 ** (-sweep_num / tau))
            if converged_num > converged_count:
                break
            # energy
            energy_at_edge = deepcopy(_energy_at_edge)
            ee_at_edge = deepcopy(_ee_at_edge)
            edges = deepcopy(_edges)

            self.distance = self.initial_distance()
            self.flag = self.initial_flag()

            (
                _edge_id,
                _selected_tensor_id,
                _connected_tensor_id,
                _not_selected_tensor_id,
            ) = self.local_two_tensor()

            print("Sweep count: " + str(sweep_num + 1))
            while True:
                edge_id = _edge_id
                selected_tensor_id = _selected_tensor_id
                connected_tensor_id = _connected_tensor_id
                not_selected_tensor_id = _not_selected_tensor_id

                # absorb gauge tensor
                iso = tn.Node(
                    self.psi.tensors[selected_tensor_id], backend=self.backend
                )
                gauge = tn.Node(self.psi.gauge_tensor, backend=self.backend)
                if selected_tensor_id == self.previous_id:
                    out = gauge[1]
                    iso[2] ^ gauge[0]
                else:
                    out = gauge[0]
                    iso[2] ^ gauge[1]
                iso = tn.contractors.auto(
                    [iso, gauge], output_edge_order=[iso[0], iso[1], out, gauge[2]]
                )

                self.psi.tensors[selected_tensor_id] = iso.tensor

                self.set_ttn_properties_at_one_tensor(edge_id, selected_tensor_id)

                self._set_edge_spin(not_selected_tensor_id)

                self._set_block_hamiltonian(not_selected_tensor_id)

                ground_state_order = [selected_tensor_id, connected_tensor_id]
                ground_state, energy = self.lanczos(ground_state_order)

                psi_edges = (
                    self.psi.edges[selected_tensor_id][:2]
                    + self.psi.edges[connected_tensor_id][:2]
                )

                u, s, v, probability, error, edge_order = self.decompose_two_tensors(
                    ground_state,
                    self.max_bond_dim,
                    opt_structure=opt_structure,
                    temperature=temp,
                    epsilon=entanglement_convergence_threshold,
                    delta=self.entanglement_degeneracy_threshold,
                )

                self.psi.tensors[selected_tensor_id] = u
                self.psi.tensors[connected_tensor_id] = v
                self.psi.gauge_tensor = s
                (
                    self.psi.edges[selected_tensor_id][0],
                    self.psi.edges[selected_tensor_id][1],
                ) = (
                    psi_edges[edge_order[0]],
                    psi_edges[edge_order[1]],
                )
                (
                    self.psi.edges[connected_tensor_id][0],
                    self.psi.edges[connected_tensor_id][1],
                ) = (
                    psi_edges[edge_order[2]],
                    psi_edges[edge_order[3]],
                )

                self.previous_id = selected_tensor_id

                self.distance = self.initial_distance()
                _energy_at_edge[self.psi.canonical_center_edge_id] = energy
                print(energy)
                ee = self.entanglement_entropy(probability)
                _ee_at_edge[self.psi.canonical_center_edge_id] = ee
                ee_dict = self.entanglement_entropy_at_physical_bond(
                    ground_state, psi_edges
                )
                for key in ee_dict.keys():
                    _ee_at_edge[key] = ee_dict[key]
                _error_at_edge[self.psi.canonical_center_edge_id] = error

                if sz_sign != 0 and self.candidate_edge_ids() == []:
                    # absorb gauge tensor
                    iso = tn.Node(
                        self.psi.tensors[selected_tensor_id], backend=self.backend
                    )
                    c0 = (
                        self.psi.tensors[selected_tensor_id]
                        .flat_charges[2]
                        .charges.flatten()
                    )
                    c1 = (
                        self.psi.tensors[connected_tensor_id]
                        .flat_charges[2]
                        .charges.flatten()
                    )
                    gauge_tensor = BlockSparseTensor.random(
                        [
                            Index(U1Charge(c0), flow=True),
                            Index(U1Charge(c1), flow=True),
                            Index(
                                U1Charge([self.init_u1_num + int(sz_sign * 2)]),
                                flow=False,
                            ),
                        ]
                    )
                    gauge_tensor /= np.linalg.norm(gauge_tensor.data)
                    gauge = tn.Node(gauge_tensor, backend=self.backend)
                    self.init_u1_num += int(sz_sign * 2)
                    out = gauge[1]
                    iso[2] ^ gauge[0]
                    iso = tn.contractors.auto(
                        [iso, gauge], output_edge_order=[iso[0], iso[1], out, gauge[2]]
                    )
                    self.psi.tensors[selected_tensor_id] = iso.tensor
                    ground_state, energy = self.lanczos(
                        [selected_tensor_id, connected_tensor_id]
                    )
                    self.psi.tensors[selected_tensor_id] = iso.tensor
                    u, s, v, probability, error, edge_order = (
                        self.decompose_two_tensors(
                            ground_state,
                            self.max_bond_dim,
                            opt_structure=False,
                            epsilon=entanglement_convergence_threshold,
                            delta=self.entanglement_degeneracy_threshold,
                        )
                    )
                    self.psi.tensors[selected_tensor_id] = u
                    self.psi.tensors[connected_tensor_id] = v
                    self.psi.gauge_tensor = s
                    (
                        self.psi.edges[selected_tensor_id][0],
                        self.psi.edges[selected_tensor_id][1],
                    ) = (
                        psi_edges[edge_order[0]],
                        psi_edges[edge_order[1]],
                    )
                    (
                        self.psi.edges[connected_tensor_id][0],
                        self.psi.edges[connected_tensor_id][1],
                    ) = (
                        psi_edges[edge_order[2]],
                        psi_edges[edge_order[3]],
                    )
                if self.candidate_edge_ids() == []:
                    break

                (
                    _edge_id,
                    _selected_tensor_id,
                    _connected_tensor_id,
                    _not_selected_tensor_id,
                ) = self.local_two_tensor()
                self.set_flag(_not_selected_tensor_id)

                # eval expval
                if self.flag[_not_selected_tensor_id]:
                    if eval_onesite_expval:
                        onesite_expval_dict = self.expval_onesite(
                            _not_selected_tensor_id,
                            ground_state,
                            ground_state_order,
                        )
                        for key in onesite_expval_dict.keys():
                            onesite_expval[key] = onesite_expval_dict[key]

                    if eval_twosite_expval:
                        twosite_expval_dict = self.expval_twosite(
                            _not_selected_tensor_id,
                            ground_state,
                            ground_state_order,
                        )
                        for key in twosite_expval_dict.keys():
                            twosite_expval[key] = twosite_expval_dict[key]

            # eval expval
            if eval_onesite_expval:
                for i in self.psi.central_tensor_ids():
                    onesite_expval_dict = self.expval_onesite(
                        i, ground_state, ground_state_order
                    )
                    for key in onesite_expval_dict.keys():
                        onesite_expval[key] = onesite_expval_dict[key]
            if eval_twosite_expval:
                for i in self.psi.central_tensor_ids():
                    twosite_expval_dict = self.expval_twosite(
                        i, ground_state, ground_state_order
                    )
                    for key in twosite_expval_dict.keys():
                        twosite_expval[key] = twosite_expval_dict[key]
                twosite_expval_dict = self.expval_twosite_origin(
                    twosite_expval.keys(), ground_state, ground_state_order
                )
                for key in twosite_expval_dict.keys():
                    twosite_expval[key] = twosite_expval_dict[key]

            _edges = deepcopy(self.psi.edges)

            sweep_num += 1
            if sweep_num > 2:
                diff_energy = [
                    np.abs(1 - _energy_at_edge[key] / energy_at_edge[key])
                    for key in energy_at_edge.keys()
                ]
                diff_ee = [
                    np.abs(ee_at_edge[key] - _ee_at_edge[key])
                    for key in ee_at_edge.keys()
                ]
                if all(
                    [
                        set(edge[:2]) == set(_edge[:2]) and edge[2] == _edge[2]
                        for edge, _edge in zip(edges, _edges)
                    ]
                ):
                    if all(
                        [
                            energy < energy_convergence_threshold
                            for energy in diff_energy
                        ]
                    ) and all(
                        [ee < entanglement_convergence_threshold for ee in diff_ee]
                    ):
                        converged_num += 1
        print("Converged")

        self.energy = _energy_at_edge
        self.entanglement = _ee_at_edge
        self.error = _error_at_edge
        self.one_site_expval = onesite_expval
        self.two_site_expval = twosite_expval

        return 0



import argparse
import os
from pathlib import Path

import numpy as np
import pandas as pd
import yaml
from dotmap import DotMap

from ttnopt.src import FactorizeTensor, TreeTensorNetwork


def factorize_tensor():
    parser = argparse.ArgumentParser(description="Factorize tensor")
    parser.add_argument(
        "config_file", type=str, help="Path to the YAML configuration file"
    )
    args = parser.parse_args()

    with open(args.config_file, "r") as file:
        config = yaml.safe_load(file)
    config = DotMap(config)
    numerics = config.numerics

    path = (
        Path(config.output.dir)
        if not isinstance(config.output.dir, DotMap)
        else Path("./")
    )
    os.makedirs(path, exist_ok=True)

    if isinstance(config.target.tensor, DotMap):
        print("=" * 50)
        print("⚠️  Error:")
        print("     Please set file or folder of the target.")
        print("=" * 50)
        exit()
    if not os.path.exists(str(config.target.tensor)):
        print("=" * 50)
        print("⚠️  Error:")
        print("     Please guarantee the file or folder of the target exist.")
        print("=" * 50)
        exit()

    save_tensors = (
        config.output.tensors
        if not isinstance(config.output.tensors, DotMap)
        else False
    )

    if os.path.isfile(str(config.target.tensor)):
        quantum_state = np.load(config.target.tensor)
        state_norm = np.linalg.norm(quantum_state)
        quantum_state = quantum_state / state_norm
        N = len(quantum_state.shape)
        psi = TreeTensorNetwork.mps(
            N,
            target=quantum_state,
            max_bond_dimension=numerics.initial_bond_dimension,
        )

        init_bond_dim = 4
        if not isinstance(numerics.initial_bond_dimension, DotMap):
            init_bond_dim = int(numerics.initial_bond_dimension)

        max_truncation_error = 0.0
        if not isinstance(numerics.max_truncation_error, DotMap):
            max_truncation_error = float(numerics.max_truncation_error)

        ft = FactorizeTensor(
            psi,
            max_bond_dim=init_bond_dim,
        )

        if not isinstance(numerics.opt_structure.type, DotMap):
            temperature = (
                float(numerics.opt_structure.temperature)
                if (
                    isinstance(numerics.opt_structure.temperature, float)
                    or isinstance(numerics.opt_structure.temperature, int)
                )
                else 0.0
            )
            seed = (
                numerics.opt_structure.seed
                if isinstance(numerics.opt_structure.seed, int)
                else 0
            )
            np.random.seed(seed)

            tau = (
                numerics.opt_structure.tau
                if isinstance(numerics.opt_structure.tau, int)
                else numerics.max_bond_dimension // 2
            )

            if numerics.opt_structure.type:
                max_num_sweep = 1
                if not isinstance(numerics.max_num_sweep, DotMap):
                    max_num_sweep = numerics.max_num_sweep
                ft.run(
                    target=quantum_state,
                    opt_fidelity=False,
                    opt_structure=numerics.opt_structure.type,
                    temperature=temperature,
                    tau=tau,
                    max_num_sweep=max_num_sweep,
                    max_truncation_error=max_truncation_error,
                )
            else:
                ft.run(
                    target=quantum_state,
                    opt_fidelity=False,
                    opt_structure=0,
                    max_num_sweep=1,
                    max_truncation_error=max_truncation_error,
                )
            nodes_list = {}
            for edge_id in ft.fidelity.keys():
                tmp = []
                for node_id, edges in enumerate(psi.edges):
                    node_id += N
                    if edge_id in edges:
                        tmp.append(node_id)
                nodes_list[edge_id] = tmp

            for edge_id in psi.physical_edges:
                tmp = []
                tmp.append(edge_id)
                for node_id, edges in enumerate(psi.edges):
                    node_id += N
                    if edge_id in edges:
                        tmp.append(node_id)
                nodes_list[edge_id] = tmp
                ft.fidelity[edge_id] = 0.0
                ft.error[edge_id] = 0.0

            all_keys = set(nodes_list.keys())
            df = pd.DataFrame(
                [nodes_list[k] for k in all_keys],
                columns=["node1", "node2"],
                index=None,
            )
            df["entanglement"] = [ft.entanglement[k] for k in all_keys]
            df["fidelity"] = [ft.fidelity[k] for k in all_keys]
            df["error"] = [ft.error[k] for k in all_keys]
            df.to_csv(path / "basic.csv", header=True, index=None)

        np.savetxt(path / "graph.dat", ft.psi.edges, fmt="%d", delimiter=",")
        if save_tensors:
            file_psi = path / "tensors_info"
            os.makedirs(file_psi, exist_ok=True)
            for i, iso in enumerate(ft.psi.tensors):
                np.save(file_psi / f"tensor{i}.npy", iso)
            np.save(file_psi / "singular_values.npy", ft.psi.gauge_tensor)
            np.save(file_psi / "norm.npy", state_norm)

        opt_fidelity = (
            True
            if not isinstance(numerics.fidelity.opt_structure.type, DotMap)
            else False
        )

        if opt_fidelity:
            opt_structure = numerics.fidelity.opt_structure.type
            temperature = (
                float(numerics.fidelity.opt_structure.temperature)
                if (
                    isinstance(numerics.fidelity.opt_structure.temperature, float)
                    or isinstance(numerics.fidelity.opt_structure.temperature, int)
                )
                else 0.0
            )
            seed = (
                numerics.opt_structure.seed
                if isinstance(numerics.fidelity.opt_structure.seed, int)
                else 0
            )
            np.random.seed(seed)

            tau = (
                numerics.fidelity.opt_structure.tau
                if isinstance(numerics.fidelity.opt_structure.tau, int)
                else numerics.fidelity.max_bond_dimensions[0] // 2
            )

            for i, (max_bond_dim, max_num_sweep) in enumerate(
                zip(
                    numerics.fidelity.max_bond_dimensions,
                    numerics.fidelity.max_num_sweeps,
                )
            ):
                ft.max_bond_dim = max_bond_dim
                ft.run(
                    target=quantum_state,
                    opt_fidelity=True,
                    opt_structure=opt_structure,
                    temperature=temperature,
                    tau=tau,
                    max_num_sweep=max_num_sweep,
                    max_truncation_error=max_truncation_error,
                )
                opt_structure = 0

                nodes_list = {}
                for edge_id in ft.fidelity.keys():
                    tmp = []
                    for node_id, edges in enumerate(psi.edges):
                        node_id += N
                        if edge_id in edges:
                            tmp.append(node_id)
                    nodes_list[edge_id] = tmp

                for edge_id in psi.physical_edges:
                    tmp = []
                    tmp.append(edge_id)
                    for node_id, edges in enumerate(psi.edges):
                        node_id += N
                        if edge_id in edges:
                            tmp.append(node_id)
                    nodes_list[edge_id] = tmp
                    ft.fidelity[edge_id] = 0.0
                    ft.error[edge_id] = 0.0
                    ft.bond_dim[edge_id] = quantum_state.shape[edge_id]

                all_keys = set(nodes_list.keys())
                df = pd.DataFrame(
                    [nodes_list[k] for k in all_keys],
                    columns=["node1", "node2"],
                    index=None,
                )
                df["entanglement"] = [ft.entanglement[k] for k in all_keys]
                df["fidelity"] = [ft.fidelity[k] for k in all_keys]
                df["error"] = [ft.error[k] for k in all_keys]
                df["bond"] = [ft.bond_dim[k] for k in all_keys]

                path_ = path / f"run{i + 1}"
                os.makedirs(path_, exist_ok=True)
                df.to_csv(path_ / "basic.csv", header=True, index=None)

                np.savetxt(
                    path_ / "graph.dat",
                    ft.psi.edges,
                    fmt="%d",
                    delimiter=",",
                )
                if save_tensors:
                    file_psi = Path(path_) / "tensors_info"
                    os.makedirs(file_psi, exist_ok=True)
                    for i, iso in enumerate(ft.psi.tensors):
                        np.save(file_psi / f"isometry{i}.npy", iso)
                    np.save(file_psi / "singular_values.npy", ft.psi.gauge_tensor)
                    np.save(file_psi / "norm.npy", state_norm)

    elif os.path.isdir(str(config.target.tensor)):
        input_path = Path(config.target.tensor)
        isometries = list(input_path.glob("isometry*.npy"))
        if isometries == []:
            raise FileNotFoundError(
                f"No files found files named as 'tensor' in {str(config.target.tensor)}"
            )
        isometries.sort(key=lambda x: int(x.stem.split("isometry")[-1]))
        isometries = [np.load(iso) for iso in isometries]
        singular_values = np.load(input_path / "singular_values.npy")
        state_norm = np.load(input_path / "norm.npy")
        graph_file = config.target.graph
        if os.path.isfile(str(config.target.graph)):
            edges = pd.read_csv(graph_file, delimiter=",", header=None).values
        else:
            print("=" * 50)
            print("⚠️  Error:")
            print("     target.graph is must be graph file.")
            print("=" * 50)
            exit()
        edges = [list(edge.tolist()) for edge in edges]
        psi = TreeTensorNetwork(
            edges, tensors=isometries, gauge_tensor=singular_values, norm=state_norm
        )
        N = len(psi.physical_edges)

        max_bond_dim = np.max([iso.shape[2] for iso in isometries])

        init_bond_dim = 4
        if not isinstance(numerics.initial_bond_dimension, DotMap):
            init_bond_dim = int(numerics.initial_bond_dimension)

        max_truncation_error = 0.0
        if not isinstance(numerics.max_truncation_error, DotMap):
            max_truncation_error = float(numerics.max_truncation_error)

        ft = FactorizeTensor(
            psi,
            max_bond_dim=max_bond_dim,
        )

        opt_structure = numerics.opt_structure.type

        temperature = (
            float(numerics.opt_structure.temperature)
            if (
                isinstance(numerics.opt_structure.temperature, float)
                or isinstance(numerics.opt_structure.temperature, int)
            )
            else 0.0
        )
        seed = (
            numerics.opt_structure.seed
            if isinstance(numerics.opt_structure.seed, int)
            else 0
        )
        np.random.seed(seed)

        tau = (
            numerics.opt_structure.tau
            if isinstance(numerics.opt_structure.tau, int)
            else numerics.max_num_sweep // 2 + 1
        )

        ft.run(
            opt_fidelity=False,
            opt_structure=opt_structure,
            temperature=temperature,
            tau=tau,
            max_num_sweep=numerics.max_num_sweep,
            max_truncation_error=max_truncation_error,
        )
        nodes_list = {}
        for edge_id in ft.error.keys():
            tmp = []
            for node_id, edges in enumerate(psi.edges):
                node_id += N
                if edge_id in edges:
                    tmp.append(node_id)
            nodes_list[edge_id] = tmp

        for edge_id in psi.physical_edges:
            tmp = []
            tmp.append(edge_id)
            for node_id, edges in enumerate(psi.edges):
                node_id += N
                if edge_id in edges:
                    tmp.append(node_id)
            nodes_list[edge_id] = tmp
            ft.error[edge_id] = 0.0
            for t, edges in enumerate(ft.psi.edges):
                if edge_id == edges[0]:
                    ft.bond_dim[edge_id] = ft.psi.tensors[t].shape[0]
                if edge_id == edges[1]:
                    ft.bond_dim[edge_id] = ft.psi.tensors[t].shape[1]

        all_keys = set(nodes_list.keys())
        df = pd.DataFrame(
            [nodes_list[k] for k in all_keys],
            columns=["node1", "node2"],
            index=None,
        )
        df["entanglement"] = [ft.entanglement[k] for k in all_keys]
        df["error"] = [ft.error[k] for k in all_keys]
        df["bond"] = [ft.bond_dim[k] for k in all_keys]

        path
        df.to_csv(path / "basic.csv", header=True, index=None)
        np.savetxt(path / "graph.dat", ft.psi.edges, fmt="%d", delimiter=",")

        if save_tensors:
            file_psi = path / "tensors_info"
            os.makedirs(file_psi, exist_ok=True)
            for i, iso in enumerate(ft.psi.tensors):
                np.save(file_psi / f"isometry{i}.npy", iso)
            np.save(file_psi / "singular_values.npy", ft.psi.gauge_tensor)
            np.save(file_psi / "norm.npy", state_norm)

    return 0



from typing import Union
import numpy as np
import tensornetwork as tn
from tensornetwork import U1Charge, Index, BlockSparseTensor
from scipy.linalg import eigh_tridiagonal
from copy import deepcopy
from collections import defaultdict

from ttnopt.src.TTN import TreeTensorNetwork
from ttnopt.src.Hamiltonian import Hamiltonian
from ttnopt.src.Observable import bare_spin_operator, spin_dof, spin_ind
from ttnopt.src.TwoSiteUpdaterSparse import TwoSiteUpdaterSparse
from ttnopt.src.functionTTN import get_renormalization_sequence, get_bare_edges


class PhysicsEngineSparse(TwoSiteUpdaterSparse):

    def __init__(
        self,
        psi: TreeTensorNetwork,
        hamiltonian: Hamiltonian,
        u1_num: Union[int, str],
        init_bond_dim: int,
        max_bond_dim: int,
        energy_degeneracy_threshold: float = 1e-13,
        entanglement_degeneracy_threshold: float = 0.1,
    ):
        """Initialize a PhysicsEngineSparse object.

        Args:
            psi (TreeTensorNetwork): The quantum state.
            hamiltonians (Hamiltonian): Hamiltonian which has a list of Observables.
            u1_num (int or str): The number of U1 charges.
            init_bond_dim (int): Initial bond dimension.
            max_bond_dim (int): Maximum bond dimension.
            edge_spin_operators (Optional(Dict[int, Dict[str, np.ndarray]]): Spin operators at each edge. Defaults to None.
        """

        super().__init__(psi)
        self.hamiltonian = hamiltonian
        self.u1_num = int(2 * spin_ind("S=" + str(u1_num)))
        self.init_u1_num = int(2 * spin_ind("S=" + str(u1_num)))
        self.init_bond_dim = init_bond_dim
        self.max_bond_dim = max_bond_dim
        self.edge_u1_charges = self._init_edge_u1_charge()
        self.edge_spin_operators = self._init_spin_operator()
        self.block_hamiltonians = self._init_block_hamiltonians()
        self.previous_id = 0
        self.energy_degeneracy_threshold = energy_degeneracy_threshold
        self.entanglement_degeneracy_threshold = entanglement_degeneracy_threshold

        init_tensors_flag = False
        if (
            self.psi.tensors is None
        ):  # if there is no initial tensors, we need to generate it
            print("No initial tensors in TTN object.")
            self.psi.tensors = []
            for _ in self.psi.edges:
                self.psi.tensors.append(None)
            init_tensors_flag = True
        else:
            for k in self.hamiltonian.spin_size.keys():
                if spin_dof(self.hamiltonian.spin_size[k]) != self.psi.edge_dims[k]:
                    print("Initial tensors is not valid for given hamiltonian.")
                    init_tensors_flag = True
                    break

        if init_tensors_flag:
            print("Initialize tensors with real space renormalization.")

            for k in self.hamiltonian.spin_size.keys():
                self.psi.edge_dims[k] = spin_dof(self.hamiltonian.spin_size[k])
            self.init_tensors_by_block_hamiltonian()

    def expval_onesite(self, tensor_id, ground_state, tensor_ids):
        """Calculate the expectation values of the one-site operators.
        Returns:
            The expectation values of the one-site operators of dict.
        """
        one_site_expvals = {}
        indices = [i for i in self.psi.edges[tensor_id][:2]]
        bra_tensor = ground_state.copy()
        start_index = 0 if tensor_id == tensor_ids[0] else 2
        ket_tensor = bra_tensor.copy(conjugate=True)
        for index in indices:
            if index in self.psi.physical_edges:
                expvals = {}
                for operator in ["Sz"]:
                    bra = bra_tensor.copy()
                    ket = ket_tensor.copy()
                    spin = tn.Node(
                        self._spin_operator_at_edge(index, index, operator),
                        backend=self.backend,
                    )
                    if index == self.psi.edges[tensor_id][0]:
                        bra[start_index + 0] ^ spin[0]
                        if start_index == 0:
                            bra = tn.contractors.auto(
                                [bra, spin],
                                output_edge_order=[
                                    spin[1],
                                    bra[1],
                                    bra[2],
                                    bra[3],
                                    bra[4],
                                ],
                            )
                        else:
                            bra = tn.contractors.auto(
                                [bra, spin],
                                output_edge_order=[
                                    bra[0],
                                    bra[1],
                                    spin[1],
                                    bra[3],
                                    bra[4],
                                ],
                            )
                    if index == self.psi.edges[tensor_id][1]:
                        bra[start_index + 1] ^ spin[0]
                        if start_index == 0:
                            bra = tn.contractors.auto(
                                [bra, spin],
                                output_edge_order=[
                                    bra[0],
                                    spin[1],
                                    bra[2],
                                    bra[3],
                                    bra[4],
                                ],
                            )
                        else:
                            bra = tn.contractors.auto(
                                [bra, spin],
                                output_edge_order=[
                                    bra[0],
                                    bra[1],
                                    bra[2],
                                    spin[1],
                                    bra[4],
                                ],
                            )
                    bra[0] ^ ket[0]
                    bra[1] ^ ket[1]
                    bra[2] ^ ket[2]
                    bra[3] ^ ket[3]
                    bra[4] ^ ket[4]
                    expvals[operator] = tn.contractors.auto(
                        [bra, ket]
                    ).tensor.data.item()
                one_site_expvals[index] = expvals
        return one_site_expvals

    def expval_twosite(self, tensor_id, ground_state, tensor_ids):
        two_site_expvals = {}
        l_bare_edges = get_bare_edges(
            self.psi.edges[tensor_id][0],
            self.psi.edges,
            self.psi.physical_edges,
        )
        r_bare_edges = get_bare_edges(
            self.psi.edges[tensor_id][1],
            self.psi.edges,
            self.psi.physical_edges,
        )
        start_index = 0 if tensor_id == tensor_ids[0] else 2
        bra_tensor = ground_state.copy()
        ket_tensor = bra_tensor.copy(conjugate=True)
        pairs = [(i, j) for i in l_bare_edges for j in r_bare_edges]
        for pair in pairs:
            expvals = {}
            for operators in [
                ["Sz", "Sz"],
                ["S+", "S-"],
                ["S-", "S+"],
            ]:
                bra = bra_tensor.copy()
                ket = ket_tensor.copy()
                spin1 = tn.Node(
                    self._spin_operator_at_edge(
                        self.psi.edges[tensor_id][0], pair[0], operators[0]
                    ),
                    backend=self.backend,
                )
                spin2 = tn.Node(
                    self._spin_operator_at_edge(
                        self.psi.edges[tensor_id][1], pair[1], operators[1]
                    ),
                    backend=self.backend,
                )
                if operators != ["Sz", "Sz"]:
                    bra_ = bra.tensor
                    charges = [
                        bra_.flat_charges[i].charges.flatten()
                        for i in range(len(bra.shape))
                    ]
                    charges[0 + start_index] = spin1.tensor.flat_charges[
                        0
                    ].charges.flatten()
                    charges[1 + start_index] = spin2.tensor.flat_charges[
                        0
                    ].charges.flatten()
                    bra_.contiguous(inplace=True)
                    bra_ = BlockSparseTensor(
                        data=bra_.data,
                        charges=[U1Charge(c) for c in charges],
                        flows=bra_.flat_flows,
                    )
                    bra = tn.Node(bra_, backend=self.backend)

                bra[start_index + 0] ^ spin1[0]
                if start_index == 0:
                    bra = tn.contractors.auto(
                        [bra, spin1],
                        output_edge_order=[spin1[1], bra[1], bra[2], bra[3], bra[4]],
                    )
                else:
                    bra = tn.contractors.auto(
                        [bra, spin1],
                        output_edge_order=[bra[0], bra[1], spin1[1], bra[3], bra[4]],
                    )
                bra[start_index + 1] ^ spin2[0]
                if start_index == 0:
                    bra = tn.contractors.auto(
                        [bra, spin2],
                        output_edge_order=[bra[0], spin2[1], bra[2], bra[3], bra[4]],
                    )
                else:
                    bra = tn.contractors.auto(
                        [bra, spin2],
                        output_edge_order=[bra[0], bra[1], bra[2], spin2[1], bra[4]],
                    )
                bra[0] ^ ket[0]
                bra[1] ^ ket[1]
                bra[2] ^ ket[2]
                bra[3] ^ ket[3]
                bra[4] ^ ket[4]
                exp_val = tn.contractors.auto([bra, ket]).tensor.data.item()
                op_key = (
                    operators[0] + operators[1]
                    if pair[0] > pair[1]
                    else operators[1] + operators[0]
                )
                expvals[op_key] = exp_val
            key = (pair[0], pair[1]) if pair[0] < pair[1] else (pair[1], pair[0])
            two_site_expvals[key] = expvals
        return two_site_expvals

    def expval_twosite_origin(self, keys, ground_state, tensor_ids):
        two_site_expvals = {}
        out_edge_dict = {}
        l_bare_edges = get_bare_edges(
            self.psi.edges[tensor_ids[0]][0],
            self.psi.edges,
            self.psi.physical_edges,
        )
        for ll in l_bare_edges:
            out_edge_dict[ll] = 0
        l_bare_edges_ = get_bare_edges(
            self.psi.edges[tensor_ids[0]][1],
            self.psi.edges,
            self.psi.physical_edges,
        )
        l_bare_edges += l_bare_edges_
        for ll in l_bare_edges_:
            out_edge_dict[ll] = 1

        r_bare_edges = get_bare_edges(
            self.psi.edges[tensor_ids[1]][0],
            self.psi.edges,
            self.psi.physical_edges,
        )
        for r in r_bare_edges:
            out_edge_dict[r] = 0
        r_bare_edges_ = get_bare_edges(
            self.psi.edges[tensor_ids[1]][1],
            self.psi.edges,
            self.psi.physical_edges,
        )
        r_bare_edges += r_bare_edges_
        for r in r_bare_edges_:
            out_edge_dict[r] = 1

        bra_tensor = ground_state.copy()
        ket_tensor = bra_tensor.copy(conjugate=True)

        pairs = [(i, j) for i in l_bare_edges for j in r_bare_edges]
        pairs = [pair for pair in pairs if tuple(sorted(pair)) not in keys]
        for pair in pairs:
            expvals = {}
            for operators in [
                ["Sz", "Sz"],
                ["S+", "S-"],
                ["S-", "S+"],
            ]:
                bra = bra_tensor.copy()
                ket = ket_tensor.copy()
                spin1 = tn.Node(
                    self._spin_operator_at_edge(
                        self.psi.edges[tensor_ids[0]][out_edge_dict[pair[0]]],
                        pair[0],
                        operators[0],
                    ),
                    backend=self.backend,
                )
                spin2 = tn.Node(
                    self._spin_operator_at_edge(
                        self.psi.edges[tensor_ids[1]][out_edge_dict[pair[1]]],
                        pair[1],
                        operators[1],
                    ),
                    backend=self.backend,
                )
                if operators != ["Sz", "Sz"]:
                    bra_ = bra.tensor
                    charges = [
                        bra_.flat_charges[i].charges.flatten()
                        for i in range(len(bra.shape))
                    ]
                    charges[out_edge_dict[pair[0]]] = spin1.tensor.flat_charges[
                        0
                    ].charges.flatten()
                    charges[out_edge_dict[pair[1]] + 2] = spin2.tensor.flat_charges[
                        0
                    ].charges.flatten()
                    bra_.contiguous(inplace=True)
                    bra_ = BlockSparseTensor(
                        data=bra_.data,
                        charges=[U1Charge(c) for c in charges],
                        flows=bra_.flat_flows,
                    )
                    bra = tn.Node(bra_, backend=self.backend)

                out = [bra[0], bra[1], bra[2], bra[3], bra[4]]
                bra[out_edge_dict[pair[0]]] ^ spin1[0]
                out[out_edge_dict[pair[0]]] = spin1[1]
                bra = tn.contractors.auto([bra, spin1], output_edge_order=out)
                bra[out_edge_dict[pair[1]] + 2] ^ spin2[0]
                out[out_edge_dict[pair[1]] + 2] = spin2[1]
                bra = tn.contractors.auto([bra, spin2], output_edge_order=out)

                bra[0] ^ ket[0]
                bra[1] ^ ket[1]
                bra[2] ^ ket[2]
                bra[3] ^ ket[3]
                bra[4] ^ ket[4]

                exp_val = tn.contractors.auto([bra, ket]).tensor.data.item()
                op_key = (
                    operators[0] + operators[1]
                    if pair[0] < pair[1]
                    else operators[1] + operators[0]
                )
                expvals[op_key] = exp_val
            key = (pair[0], pair[1]) if pair[0] < pair[1] else (pair[1], pair[0])
            two_site_expvals[key] = expvals
        return two_site_expvals

    def lanczos(
        self,
        central_tensor_ids,
        lanczos_tol=1e-13,
        inverse_tol=1e-7,
    ):
        psi_1 = tn.Node(self.psi.tensors[central_tensor_ids[0]], backend=self.backend)
        psi_2 = tn.Node(self.psi.tensors[central_tensor_ids[1]], backend=self.backend)
        psi_1[2] ^ psi_2[2]
        psi = tn.contractors.auto(
            [psi_1, psi_2],
            output_edge_order=[psi_1[0], psi_1[1], psi_2[0], psi_2[1], psi_1[3]],
        )
        # normalization
        psi = psi / np.linalg.norm(psi.tensor.data)
        psi_ = psi.copy()
        psi_0 = psi.copy()
        dim_n = np.prod(psi.shape)
        alpha = np.zeros(dim_n, dtype=np.float64)
        beta = np.zeros(dim_n, dtype=np.float64)

        psi_w = self._apply_ham_psi(psi, central_tensor_ids)

        alpha[0] = np.real(np.dot(psi_w.tensor.data, psi.tensor.data))
        omega = psi_w.tensor - alpha[0] * psi.tensor

        d = 0
        if dim_n == 1:
            print("-" * 50)
            print("Fail on lanczos: All bond dimensions in canonical center are 1.")
            print(
                "Set more larger bond dimension to run correctly on numerics.max_bond_dimensions."
            )
            print("-" * 50)
            exit()
        else:
            e_old = 0.0
            for j in range(1, dim_n):
                beta[j] = np.linalg.norm(omega.data)
                psi = tn.Node(omega / beta[j], backend=self.backend)
                psi_w = self._apply_ham_psi(psi, central_tensor_ids)
                alpha[j] = np.real(np.dot(psi_w.tensor.data, psi.tensor.data))
                omega = psi_w.tensor - alpha[j] * psi.tensor - beta[j] * psi_.tensor
                psi_ = psi

                if j >= 1:
                    if j > dim_n:
                        break
                    e, v_tilda = eigh_tridiagonal(
                        np.real(alpha[: j + 1]),
                        np.real(beta[1 : j + 1]),
                        select="i",
                        select_range=(0, 0),
                    )
                    energy = e[0]
                    if np.abs(e - e_old) < np.max([1.0, np.abs(e)[0]]) * lanczos_tol:
                        d += 1
                    if j > dim_n or d > 5:
                        max_e, _ = eigh_tridiagonal(
                            np.real(alpha[: j + 1]),
                            np.real(beta[1 : j + 1]),
                            select="a",
                        )
                        max_e = max_e[-1]
                        break
                    e_old = energy

        v_tilda = np.array(v_tilda.flatten(), dtype=np.complex128)
        v = v_tilda[0] * psi_0.tensor
        psi = psi_0
        psi_ = psi_0
        psi_w = self._apply_ham_psi(psi, central_tensor_ids)
        a = np.real(np.dot(psi_w.tensor.data, psi.tensor.data))
        omega = psi_w.tensor - a * psi.tensor
        for k in range(1, len(v_tilda)):
            b = np.linalg.norm(omega.data)
            psi = tn.Node(omega / b, backend=self.backend)
            v += v_tilda[k] * psi.tensor
            psi_w = self._apply_ham_psi(psi, central_tensor_ids)
            a = np.real(np.dot(psi_w.tensor.data, psi.tensor.data))
            omega = psi_w.tensor - a * psi.tensor - b * psi_.tensor
            psi_ = psi

        # check convergence
        v = tn.Node(v, backend=self.backend)
        v_ = self._apply_ham_psi(v, central_tensor_ids)
        e = np.real(np.dot(v_.tensor.data, v.tensor.data)) / np.real(
            np.linalg.norm(v.tensor.data) ** 2
        )
        delta_v = v_.tensor - max_e * v.tensor
        v_tensor = v_.tensor / np.linalg.norm(v_.tensor.data)
        v = tn.Node(v_tensor, backend=self.backend)
        while np.linalg.norm(delta_v.data) > inverse_tol:
            v_ = self._apply_ham_psi(v, central_tensor_ids)
            e = np.real(np.dot(v_.tensor.data, v.tensor.data)) / np.real(
                np.linalg.norm(v.tensor.data) ** 2
            )
            delta_v = v_.tensor - e * v.tensor
            v_tensor = v_.tensor - max_e * v.tensor
            v_tensor = v_tensor / np.linalg.norm(v_tensor.data)
            v = tn.Node(v_tensor, backend=self.backend)

        eigen_vectors = v
        energy = e
        return eigen_vectors, energy

    def init_tensors_by_block_hamiltonian(self):
        sequence = get_renormalization_sequence(self.psi.edges, self.psi.top_edge_id)
        for tensor_id in sequence:
            ham = self._get_block_hamiltonian(tensor_id)
            self._set_psi_tensor_with_ham(tensor_id, ham)
            self._set_psi_edge_dim(tensor_id)
            self._set_edge_u1_charge(tensor_id)
            self._set_edge_spin(tensor_id)
            self._set_block_hamiltonian(tensor_id)

        # gauge_tensor
        (_, selected_tensor_id, _, not_selected_tensor_id) = self.local_two_tensor()
        c0 = self.psi.tensors[selected_tensor_id].flat_charges[2].charges.flatten()
        c1 = self.psi.tensors[not_selected_tensor_id].flat_charges[2].charges.flatten()
        u = U1Charge.fuse(c0, c1)
        cc = np.count_nonzero(u == self.init_u1_num)
        if cc == 0:
            if not (np.all(u % 2 == 0) and self.u1_num % 2 == 0) and not (
                np.all(u % 2 != 0) and self.u1_num % 2 != 0
            ):
                print("-" * 50)
                print(f"Fail on RG: There is no sector with M={self.u1_num}/2.")
                print("Because of the U1 charge conservation.")
                print("To rifer see U1 charges sectors on canonical center following:")
                unique_values, counts = np.unique(u, return_counts=True)
                for val, count in zip(unique_values, counts):
                    print(f"S: {val}/2, Count: {count}")
                print("-" * 50)
                exit()
            else:
                print("-" * 50)
                print(f"Note on RG: There is no sector with M={self.u1_num}/2.")
                u_even = u[u % 2 == 0]
                u_odd = u[u % 2 != 0]
                closest_sector = self.u1_num
                if self.u1_num % 2 == 0:
                    closest_sector = min(u_even, key=lambda x: abs(x - self.u1_num))
                if self.u1_num % 2 != 0:
                    closest_sector = min(u_odd, key=lambda x: abs(x - self.u1_num))
                print(
                    f"We will start warmup sweeps from the closest sector M={closest_sector}/2."
                )
                print(
                    f"it requires {abs(closest_sector - self.u1_num) // 2}-time warmup sweep."
                )
                print("-" * 50)
                self.init_u1_num = closest_sector
                cc = np.count_nonzero(u == self.init_u1_num)
        c0 = U1Charge(c0)
        c1 = U1Charge(c1)
        c = U1Charge([self.init_u1_num])
        self.psi.gauge_tensor = BlockSparseTensor(
            data=np.ones(cc) / np.sum(np.ones(cc)),
            charges=[c0, c1, c],
            flows=[True, True, False],
        )
        self.psi.gauge_tensor = self.psi.gauge_tensor / np.linalg.norm(
            self.psi.gauge_tensor.data
        )
        iso = tn.Node(self.psi.tensors[selected_tensor_id], backend=self.backend)
        gauge = tn.Node(self.psi.gauge_tensor, backend=self.backend)
        out = gauge[1]
        iso[2] ^ gauge[0]
        iso = tn.contractors.auto(
            [iso, gauge], output_edge_order=[iso[0], iso[1], out, gauge[2]]
        )
        self.psi.tensors[selected_tensor_id] = iso.tensor
        ground_state, _ = self.lanczos([selected_tensor_id, not_selected_tensor_id])
        u, s, v, _, _, _ = self.decompose_two_tensors(
            ground_state,
            self.max_bond_dim,
            delta=self.entanglement_degeneracy_threshold,
        )
        self.psi.tensors[selected_tensor_id] = u
        self.psi.tensors[not_selected_tensor_id] = v
        self.psi.gauge_tensor = s
        self.previous_id = selected_tensor_id

    def _apply_ham_psi(self, psi, central_tensor_ids):
        indices = [
            Index(psi.tensor.flat_charges[i], flow=psi.tensor.flat_flows[i])
            for i in range(len(psi.shape))
        ]
        psi_tensor = BlockSparseTensor.zeros(indices, dtype=np.complex128)

        if self.psi.edges[central_tensor_ids[0]][0] in self.block_hamiltonians.keys():
            psi_tensor += self._block_ham_psi(
                psi, self.psi.edges[central_tensor_ids[0]][0], 0
            )

        if self.psi.edges[central_tensor_ids[0]][1] in self.block_hamiltonians.keys():
            psi_tensor += self._block_ham_psi(
                psi, self.psi.edges[central_tensor_ids[0]][1], 1
            )

        if self.psi.edges[central_tensor_ids[1]][0] in self.block_hamiltonians.keys():
            psi_tensor += self._block_ham_psi(
                psi, self.psi.edges[central_tensor_ids[1]][0], 2
            )

        if self.psi.edges[central_tensor_ids[1]][1] in self.block_hamiltonians.keys():
            psi_tensor += self._block_ham_psi(
                psi, self.psi.edges[central_tensor_ids[1]][1], 3
            )

        psi_tensor += self._ham_psi(
            psi, self.psi.edges[central_tensor_ids[0]][:2], [0, 1]
        )

        psi_tensor += self._ham_psi(
            psi,
            self.psi.edges[central_tensor_ids[1]][:2],
            [2, 3],
        )
        psi_tensor += self._ham_psi(
            psi,
            [
                self.psi.edges[central_tensor_ids[0]][0],
                self.psi.edges[central_tensor_ids[1]][0],
            ],
            [0, 2],
        )
        psi_tensor += self._ham_psi(
            psi,
            [
                self.psi.edges[central_tensor_ids[0]][0],
                self.psi.edges[central_tensor_ids[1]][1],
            ],
            [0, 3],
        )
        psi_tensor += self._ham_psi(
            psi,
            [
                self.psi.edges[central_tensor_ids[0]][1],
                self.psi.edges[central_tensor_ids[1]][0],
            ],
            [1, 2],
        )
        psi_tensor += self._ham_psi(
            psi,
            [
                self.psi.edges[central_tensor_ids[0]][1],
                self.psi.edges[central_tensor_ids[1]][1],
            ],
            [1, 3],
        )
        return tn.Node(psi_tensor, backend=self.backend)

    def _block_ham_psi(self, psi, edge_id, apply_id):
        h = tn.Node(self.block_hamiltonians[edge_id], backend=self.backend)
        psi_ = psi.copy()
        psi_[apply_id] ^ h[0]

        output_edge_order = psi_.get_all_edges()
        output_edge_order[apply_id] = h[1]

        psi_tensor = tn.contractors.auto(
            [psi_, h], output_edge_order=output_edge_order
        ).get_tensor()
        return psi_tensor

    def _ham_psi(self, psi, edge_ids, apply_ids):
        def get_psi_tensor(psi, spins, other_spins, keys, apply_ids):
            indices = [
                Index(psi.tensor.flat_charges[i], flow=psi.tensor.flat_flows[i])
                for i in range(len(psi.shape))
            ]
            psi_tensor = BlockSparseTensor.zeros(indices, dtype=np.complex128)
            for pair1, pair2 in keys:
                psi_ = psi.copy().tensor
                spin1 = spins[pair1][pair2]
                spin2 = other_spins[pair1][pair2]
                charges = [
                    psi_.flat_charges[i].charges.flatten()
                    for i in range(len(psi_.shape))
                ]
                if pair2 != "Sz":
                    psi_.contiguous(inplace=True)
                    charges[apply_ids[0]] = spin1.flat_charges[0].charges.flatten()
                    charges[apply_ids[1]] = spin2.flat_charges[0].charges.flatten()
                    psi_ = BlockSparseTensor(
                        data=psi_.data,
                        charges=[U1Charge(c) for c in charges],
                        flows=psi_.flat_flows,
                    )

                psi_ = tn.Node(psi_, backend=self.backend)

                spin_op1 = tn.Node(spin1, backend=self.backend)
                psi_[apply_ids[0]] ^ spin_op1[0]
                output_edge_order = psi_.get_all_edges()
                output_edge_order[apply_ids[0]] = spin_op1[1]
                psi_ = tn.contractors.auto(
                    [psi_, spin_op1], output_edge_order=output_edge_order
                )

                spin_op2 = tn.Node(spin2, backend=self.backend)
                psi_[apply_ids[1]] ^ spin_op2[0]
                output_edge_order = psi_.get_all_edges()
                output_edge_order[apply_ids[1]] = spin_op2[1]
                psi_ = tn.contractors.auto(
                    [psi_, spin_op2], output_edge_order=output_edge_order
                )
                psi_tensor += psi_.tensor
            return psi_tensor

        l_bare_edges = get_bare_edges(
            edge_ids[0],
            self.psi.edges,
            self.psi.physical_edges,
        )
        r_bare_edges = get_bare_edges(
            edge_ids[1],
            self.psi.edges,
            self.psi.physical_edges,
        )

        spins = defaultdict(lambda: defaultdict(lambda: None))
        other_spins = defaultdict(lambda: defaultdict(lambda: None))
        keys = []
        if len(l_bare_edges) > len(r_bare_edges):
            for ham in self.hamiltonian.observables:
                if len(ham.indices) == 2:
                    if (
                        ham.indices[0] in l_bare_edges
                        and ham.indices[1] in r_bare_edges
                    ):
                        for i, op_list in enumerate(ham.operators_list):
                            key = (
                                (ham.indices[1], op_list[1]),
                                (op_list[0]),
                            )
                            if spins[key[0]][key[1]] is None:
                                keys.append(key)
                                spins[key[0]][key[1]] = ham.coef_list[i] * deepcopy(
                                    self._spin_operator_at_edge(
                                        edge_ids[0], ham.indices[0], op_list[0]
                                    )
                                )
                                other_spins[key[0]][key[1]] = (
                                    self._spin_operator_at_edge(
                                        edge_ids[1], ham.indices[1], op_list[1]
                                    )
                                )
                            else:
                                spins[key[0]][key[1]] += ham.coef_list[i] * deepcopy(
                                    self._spin_operator_at_edge(
                                        edge_ids[0], ham.indices[0], op_list[0]
                                    )
                                )
                    elif (
                        ham.indices[1] in l_bare_edges
                        and ham.indices[0] in r_bare_edges
                    ):
                        for i, op_list in enumerate(ham.operators_list):
                            key = (
                                (ham.indices[0], op_list[0]),
                                (op_list[1]),
                            )
                            if spins[key[0]][key[1]] is None:
                                keys.append(key)
                                spins[key[0]][key[1]] = ham.coef_list[i] * deepcopy(
                                    self._spin_operator_at_edge(
                                        edge_ids[0], ham.indices[1], op_list[1]
                                    )
                                )
                                other_spins[key[0]][key[1]] = (
                                    self._spin_operator_at_edge(
                                        edge_ids[1], ham.indices[0], op_list[0]
                                    )
                                )
                            else:
                                spins[key[0]][key[1]] += ham.coef_list[i] * deepcopy(
                                    self._spin_operator_at_edge(
                                        edge_ids[0], ham.indices[1], op_list[1]
                                    )
                                )
            psi_tensor = get_psi_tensor(psi, spins, other_spins, keys, apply_ids)
            return psi_tensor
        else:
            for ham in self.hamiltonian.observables:
                if len(ham.indices) == 2:
                    if (
                        ham.indices[0] in l_bare_edges
                        and ham.indices[1] in r_bare_edges
                    ):
                        for i, op_list in enumerate(ham.operators_list):
                            key = (
                                (ham.indices[0], op_list[0]),
                                (op_list[1]),
                            )
                            if spins[key[0]][key[1]] is None:
                                keys.append(key)
                                spins[key[0]][key[1]] = ham.coef_list[i] * deepcopy(
                                    self._spin_operator_at_edge(
                                        edge_ids[1], ham.indices[1], op_list[1]
                                    )
                                )
                                other_spins[key[0]][key[1]] = (
                                    self._spin_operator_at_edge(
                                        edge_ids[0], ham.indices[0], op_list[0]
                                    )
                                )
                            else:
                                spins[key[0]][key[1]] += ham.coef_list[i] * deepcopy(
                                    self._spin_operator_at_edge(
                                        edge_ids[1], ham.indices[1], op_list[1]
                                    )
                                )
                    elif (
                        ham.indices[1] in l_bare_edges
                        and ham.indices[0] in r_bare_edges
                    ):
                        for i, op_list in enumerate(ham.operators_list):
                            key = (
                                (ham.indices[1], op_list[1]),
                                (op_list[0]),
                            )
                            if spins[key[0]][key[1]] is None:
                                keys.append(key)
                                spins[key[0]][key[1]] = ham.coef_list[i] * deepcopy(
                                    self._spin_operator_at_edge(
                                        edge_ids[1], ham.indices[0], op_list[0]
                                    )
                                )
                                other_spins[key[0]][key[1]] = (
                                    self._spin_operator_at_edge(
                                        edge_ids[0], ham.indices[1], op_list[1]
                                    )
                                )
                            else:
                                spins[key[0]][key[1]] += ham.coef_list[i] * deepcopy(
                                    self._spin_operator_at_edge(
                                        edge_ids[1], ham.indices[0], op_list[0]
                                    )
                                )
            psi_tensor = get_psi_tensor(psi, spins, other_spins, keys, apply_ids[::-1])
            return psi_tensor

    def _get_block_hamiltonian(self, tensor_id):
        def fuse_ham(block_ham, u1_charges):
            dims = block_ham.shape[: len(block_ham.shape) // 2]
            dim = np.prod(dims)
            u = U1Charge(U1Charge.fuse(u1_charges[0], u1_charges[1]))
            block_ham.reshape([dim] * 2)
            block_ham.contiguous(inplace=True)
            block_ham = BlockSparseTensor(
                data=block_ham.data,
                charges=[u, u],
                flows=[False, True],
            )
            return block_ham

        block_hams = []
        edge_ids = self.psi.edges[tensor_id][:2]
        l_bare_edges = get_bare_edges(
            edge_ids[0],
            self.psi.edges,
            self.psi.physical_edges,
        )
        r_bare_edges = get_bare_edges(
            edge_ids[1],
            self.psi.edges,
            self.psi.physical_edges,
        )
        if self.psi.tensors[tensor_id] is None:
            u1_charges = [
                self.edge_u1_charges[self.psi.edges[tensor_id][0]],
                self.edge_u1_charges[self.psi.edges[tensor_id][1]],
            ]
        else:
            u1_charges = [
                self.psi.tensors[tensor_id].flat_charges[0],
                self.psi.tensors[tensor_id].flat_charges[1],
            ]
        for ham in self.hamiltonian.observables:
            spin_operators = [None, None]
            if len(ham.indices) == 2:
                if ham.indices[0] in l_bare_edges and ham.indices[1] in r_bare_edges:
                    for n in range(ham.operators_num):
                        operators = ham.operators_list[n]
                        spin_operators[0] = deepcopy(
                            self._spin_operator_at_edge(
                                edge_ids[0], ham.indices[0], operators[0]
                            )
                        )
                        spin_operators[1] = deepcopy(
                            self._spin_operator_at_edge(
                                edge_ids[1], ham.indices[1], operators[1]
                            )
                        )
                        block_ham = tn.ncon(
                            spin_operators,
                            [["-b0", "-k0"], ["-b1", "-k1"]],
                            out_order=["-b0", "-b1", "-k0", "-k1"],
                            backend=self.backend,
                        )
                        block_ham *= complex(ham.coef_list[n])
                        block_ham = fuse_ham(block_ham, u1_charges)
                        block_hams.append(block_ham)

                if ham.indices[1] in l_bare_edges and ham.indices[0] in r_bare_edges:
                    for n in range(ham.operators_num):
                        operators = ham.operators_list[n]

                        spin_operators[0] = deepcopy(
                            self._spin_operator_at_edge(
                                edge_ids[0], ham.indices[1], operators[1]
                            )
                        )
                        spin_operators[1] = deepcopy(
                            self._spin_operator_at_edge(
                                edge_ids[1], ham.indices[0], operators[0]
                            )
                        )

                        block_ham = tn.ncon(
                            spin_operators,
                            [["-b0", "-k0"], ["-b1", "-k1"]],
                            out_order=["-b0", "-b1", "-k0", "-k1"],
                            backend=self.backend,
                        )
                        block_ham *= ham.coef_list[n]
                        block_ham = fuse_ham(block_ham, u1_charges)
                        block_hams.append(block_ham)

        if self.psi.tensors[tensor_id] is None:
            b_l = U1Charge(self.edge_u1_charges[self.psi.edges[tensor_id][0]])
            b_r = U1Charge(self.edge_u1_charges[self.psi.edges[tensor_id][1]])
        else:
            b_l = U1Charge(self.psi.tensors[tensor_id].flat_charges[0])
            b_r = U1Charge(self.psi.tensors[tensor_id].flat_charges[1])

        b_l = Index(b_l, flow=False)
        b_r = Index(b_r, flow=False)
        k_l = b_l.copy().flip_flow()
        k_r = b_r.copy().flip_flow()

        eye_l = np.eye(
            self.psi.edge_dims[self.psi.edges[tensor_id][0]], dtype=np.complex128
        )

        eye_l = BlockSparseTensor.fromdense([b_l, k_l], eye_l)
        eye_r = np.eye(
            self.psi.edge_dims[self.psi.edges[tensor_id][1]], dtype=np.complex128
        )
        eye_r = BlockSparseTensor.fromdense([b_r, k_r], eye_r)

        # left block ham
        if self.psi.edges[tensor_id][0] in self.block_hamiltonians.keys():
            block_ham_left = self.block_hamiltonians[self.psi.edges[tensor_id][0]]
            block_ham = tn.ncon(
                [block_ham_left, eye_r],
                [["-b0", "-k0"], ["-b1", "-k1"]],
                out_order=["-b0", "-b1", "-k0", "-k1"],
                backend=self.backend,
            )
            block_ham = fuse_ham(block_ham, u1_charges)
            block_hams.append(block_ham)
        # right block ham
        if self.psi.edges[tensor_id][1] in self.block_hamiltonians.keys():
            block_ham_right = self.block_hamiltonians[self.psi.edges[tensor_id][1]]
            block_ham = tn.ncon(
                [eye_l, block_ham_right],
                [["-b0", "-k0"], ["-b1", "-k1"]],
                out_order=["-b0", "-b1", "-k0", "-k1"],
                backend=self.backend,
            )
            block_ham = fuse_ham(block_ham, u1_charges)
            block_hams.append(block_ham)

        # if there is no hamiltonian within this block
        if block_hams == []:
            block_ham = tn.ncon(
                [eye_l, eye_r],
                [["-b0", "-k0"], ["-b1", "-k1"]],
                out_order=["-b0", "-b1", "-k0", "-k1"],
                backend=self.backend,
            )
            block_ham = fuse_ham(block_ham, u1_charges)
            block_hams.append(block_ham)

        block_hams = np.sum(block_hams, axis=0)
        return block_hams

    def _set_psi_tensor_with_ham(self, tensor_id, ham):
        if self.psi.tensors[tensor_id] is None:
            lower_edge_dims = [
                len(self.edge_u1_charges[i]) for i in self.psi.edges[tensor_id][:2]
            ]
        else:
            lower_edge_dims = [
                len(i) for i in self.psi.tensors[tensor_id].flat_charges[:2]
            ]
        bond_dim = ham.shape[0]
        eta, iso = tn.block_sparse.eigh(ham)
        ind = np.min([self.init_bond_dim, bond_dim])
        eigenvalues = eta.data
        indices = np.argsort(eigenvalues)
        if ind < len(eigenvalues):
            while ind > 1:
                if (
                    np.abs(eigenvalues[indices[ind]] - eigenvalues[indices[ind - 1]])
                    < self.energy_degeneracy_threshold
                ):
                    ind -= 1
                else:
                    break
        selected_indices = indices[:ind]
        selected_eigenvectors = iso.todense()[:, selected_indices]
        charges = eta.flat_charges[0].charges.flatten()
        charges = charges[selected_indices]
        sorted_by_charges = np.argsort(charges)

        final_eigenvectors = selected_eigenvectors[:, sorted_by_charges]
        final_eigenvectors = final_eigenvectors.reshape(lower_edge_dims + [ind])

        charges = U1Charge(charges[sorted_by_charges])
        c0 = U1Charge(self.edge_u1_charges[self.psi.edges[tensor_id][0]])
        c1 = U1Charge(self.edge_u1_charges[self.psi.edges[tensor_id][1]])

        i = Index(charges, flow=False)
        i0 = Index(c0, flow=True)
        i1 = Index(c1, flow=True)

        self.psi.tensors[tensor_id] = BlockSparseTensor.fromdense(
            [i0, i1, i], final_eigenvectors
        )
        self.psi.tensors[tensor_id].contiguous(inplace=True)

    def _set_edge_u1_charge(self, tensor_id):
        if self.psi.tensors[tensor_id] is not None:
            self.edge_u1_charges[self.psi.edges[tensor_id][2]] = (
                self.psi.tensors[tensor_id].flat_charges[2].charges.flatten()
            )

    def _set_psi_edge_dim(self, tensor_id):
        if self.psi.tensors[tensor_id] is not None:
            self.psi.edge_dims[self.psi.edges[tensor_id][2]] = self.psi.tensors[
                tensor_id
            ].shape[2]

    def _set_edge_spin(self, tensor_id, save=True):
        new_spin_operators = {}
        # left edge
        edge_left = self.psi.edges[tensor_id][0]
        edge_right = self.psi.edges[tensor_id][1]

        bare_edges = get_bare_edges(edge_left, self.psi.edges, self.psi.physical_edges)
        spin_operators = self.edge_spin_operators[edge_left]
        for bare_edge in bare_edges:
            renormalized_spin_operators = {}
            for key, value in spin_operators[bare_edge].items():
                spin = tn.Node(value, backend=self.backend)
                bra = tn.Node(self.psi.tensors[tensor_id], backend=self.backend)
                ket = bra.copy(conjugate=True)
                bra_tensor = bra.tensor
                charges = [
                    bra_tensor.flat_charges[i].charges.flatten()
                    for i in range(len(bra.shape))
                ]
                charges[0] = spin.tensor.flat_charges[0].charges.flatten()
                if key == "S+":
                    charges[2] = charges[2] + 2
                bra_tensor.contiguous(inplace=True)
                bra = tn.Node(
                    BlockSparseTensor(
                        data=bra_tensor.data,
                        charges=[U1Charge(c) for c in charges],
                        flows=bra_tensor.flat_flows,
                    ),
                    backend=self.backend,
                )
                bra[0] ^ spin[0]
                ket[0] ^ spin[1]
                bra[1] ^ ket[1]
                spin = tn.contractors.auto(
                    [bra, spin, ket], output_edge_order=[bra[2], ket[2]]
                )
                renormalized_spin_operators[key] = spin.tensor
            new_spin_operators[bare_edge] = renormalized_spin_operators
        # right edge
        bare_edges = get_bare_edges(edge_right, self.psi.edges, self.psi.physical_edges)
        spin_operators = self.edge_spin_operators[edge_right]
        for bare_edge in bare_edges:
            renormalized_spin_operators = {}
            for key, value in spin_operators[bare_edge].items():
                spin = tn.Node(value, backend=self.backend)
                bra = tn.Node(self.psi.tensors[tensor_id], backend=self.backend)
                ket = bra.copy(conjugate=True)
                charges = [
                    bra.tensor.flat_charges[i].charges.flatten()
                    for i in range(len(bra.shape))
                ]
                bra_tensor = bra.tensor
                charges = [
                    bra_tensor.flat_charges[i].charges.flatten()
                    for i in range(len(bra.shape))
                ]
                charges[1] = spin.tensor.flat_charges[0].charges.flatten()
                if key == "S+":
                    charges[2] = charges[2] + 2
                bra_tensor.contiguous(inplace=True)
                bra = tn.Node(
                    BlockSparseTensor(
                        data=bra_tensor.data,
                        charges=[U1Charge(c) for c in charges],
                        flows=bra_tensor.flat_flows,
                    ),
                    backend=self.backend,
                )
                bra[1] ^ spin[0]
                ket[1] ^ spin[1]
                bra[0] ^ ket[0]
                spin = tn.contractors.auto(
                    [bra, spin, ket], output_edge_order=[bra[2], ket[2]]
                )
                renormalized_spin_operators[key] = spin.tensor
            new_spin_operators[bare_edge] = renormalized_spin_operators
        if save:
            self.edge_spin_operators[self.psi.edges[tensor_id][2]] = new_spin_operators
        else:
            return new_spin_operators

    def _set_block_hamiltonian(self, tensor_id):
        bra = self.psi.tensors[tensor_id]
        bra_tensor = BlockSparseTensor.zeros(
            [
                Index(bra.flat_charges[0], flow=bra.flat_flows[0]),
                Index(bra.flat_charges[1], flow=bra.flat_flows[1]),
                Index(bra.flat_charges[2], flow=bra.flat_flows[2]),
            ]
        )
        bra = tn.Node(bra, backend=self.backend)
        ket = bra.copy(conjugate=True)
        if self.psi.edges[tensor_id][0] in self.block_hamiltonians.keys():
            bra_tensor += self._block_ham_psi(bra, self.psi.edges[tensor_id][0], 0)
        if self.psi.edges[tensor_id][1] in self.block_hamiltonians.keys():
            bra_tensor += self._block_ham_psi(bra, self.psi.edges[tensor_id][1], 1)
        bra_tensor += self._ham_psi(bra, self.psi.edges[tensor_id][:2], [0, 1])

        bra_h = tn.Node(bra_tensor, backend=self.backend)
        bra_h[0] ^ ket[0]
        bra_h[1] ^ ket[1]
        block_ham = tn.contractors.auto(
            [bra_h, ket], output_edge_order=[bra_h[2], ket[2]]
        )
        self.block_hamiltonians[self.psi.edges[tensor_id][2]] = block_ham.get_tensor()

    def _spin_operator_at_edge(self, edge_id, bare_edge_id, operator):
        if operator == "S+":
            op = self.edge_spin_operators[edge_id][bare_edge_id]["S+"]
        elif operator == "S-":
            op = self.edge_spin_operators[edge_id][bare_edge_id]["S+"]
            op = op.H
            op.contiguous(inplace=True)
            op = BlockSparseTensor(
                data=op.data,
                charges=[U1Charge(c.charges.flatten() - 2) for c in op.flat_charges],
                flows=np.logical_not(op.flat_flows),
            )
        elif operator == "Sz":
            op = self.edge_spin_operators[edge_id][bare_edge_id]["Sz"]
        elif operator == "Sz2":
            op = self.edge_spin_operators[edge_id][bare_edge_id]["Sz"]
            op = BlockSparseTensor(
                data=np.array(op.data) ** 2,
                charges=op.flat_charges,
                flows=op.flat_flows,
            )
        return op

    def _init_spin_operator(self):
        edge_spin_operators = {}
        for key, value in self.hamiltonian.spin_size.items():
            z_charge = U1Charge(self.edge_u1_charges[key])
            plus_charge = U1Charge(np.array(self.edge_u1_charges[key]) + 2)
            edge_spin_operators[key] = {
                key: {
                    "Sz": BlockSparseTensor.fromdense(
                        [Index(z_charge, flow=False), Index(z_charge, flow=True)],
                        bare_spin_operator("Sz", value),
                    ),
                    "S+": BlockSparseTensor.fromdense(
                        [Index(plus_charge, flow=False), Index(z_charge, flow=True)],
                        bare_spin_operator("S+", value),
                    ),
                }
            }
        return edge_spin_operators

    def _init_block_hamiltonians(self):
        block_hamiltonians = {}
        for ham in self.hamiltonian.observables:
            for key in self.hamiltonian.spin_size.keys():
                if np.array_equal(ham.indices, [key]):
                    spin_operators = []
                    for n in range(ham.operators_num):
                        operators = ham.operators_list[n]
                        spin_operator = deepcopy(
                            self._spin_operator_at_edge(key, key, operators[0])
                        )
                        spin_operator *= ham.coef_list[n]

                        spin_operators.append(spin_operator)
                    block_ham = np.sum(spin_operators, axis=0)
                    if block_hamiltonians.get(key) is None:
                        block_hamiltonians[key] = block_ham
                    else:
                        block_hamiltonians[key] += block_ham
        return block_hamiltonians

    def _init_edge_u1_charge(self):
        edge_u1_charges = {}
        for key, spin in self.hamiltonian.spin_size.items():
            spin_num = spin_dof(spin)
            spin_value = spin_ind(spin)
            s = int(2 * spin_value)
            edge_u1_charges[key] = [-s + int(i * 2) for i in range(spin_num)]
        return edge_u1_charges

    def contract_central_tensors(self):
        central_tensor_ids = self.psi.central_tensor_ids()

        psi1 = tn.Node(self.psi.tensors[central_tensor_ids[0]], backend=self.backend)
        psi2 = tn.Node(self.psi.tensors[central_tensor_ids[1]], backend=self.backend)
        gauge = tn.Node(self.psi.gauge_tensor, backend=self.backend)

        if psi1.tensor.flows[2][0] and not psi2.tensor.flows[2][0]:
            psi1[2] ^ gauge[0]
            gauge[1] ^ psi2[2]
        if not psi1.tensor.flows[2][0] and psi2.tensor.flows[2][0]:
            psi2[2] ^ gauge[0]
            gauge[1] ^ psi1[2]

        psi = tn.contractors.auto(
            [psi1, gauge, psi2],
            output_edge_order=[psi1[0], psi1[1], psi2[0], psi2[1], gauge[2]],
        )
        return psi



import numpy as np
import tensornetwork as tn
import itertools
from collections import deque, defaultdict


class TwoSiteUpdaterMixin:
    def initial_flag(self):
        edge_ids = set(itertools.chain.from_iterable(self.psi.edges))
        flag = {ind: 0 if ind not in self.psi.physical_edges else 1 for ind in edge_ids}
        return flag

    def entanglement_entropy(self, probability):
        el = probability**2 / np.sum(probability**2)
        el = el[el > 0.0]
        ee = -np.sum(el * np.log(el))
        return np.real(ee)

    def initial_distance(self):
        adjacency_list = defaultdict(list)
        for node in self.psi.edges:
            child1, child2, parent = node
            adjacency_list[child1].append(parent)
            adjacency_list[parent].append(child1)
            adjacency_list[child2].append(parent)
            adjacency_list[parent].append(child2)
            adjacency_list[child1].append(child2)
            adjacency_list[child2].append(child1)
        for key, val in adjacency_list.items():
            adjacency_list[key] = list(set(val))

        distances = {self.psi.top_edge_id: 0}
        queue = deque([self.psi.top_edge_id])

        while queue:
            current_edge = queue.popleft()
            current_distance = distances[current_edge]
            for neighbor in adjacency_list[current_edge]:
                if neighbor not in distances.keys():
                    distances[neighbor] = current_distance + 1
                    queue.append(neighbor)
        return distances

    def candidate_edge_ids(self):
        child_tensor_ids = [
            i
            for i, edge in enumerate(self.psi.edges)
            if edge[2] == self.psi.canonical_center_edge_id
        ]
        candidate_edge_ids = (
            self.psi.edges[child_tensor_ids[0]][:2]
            + self.psi.edges[child_tensor_ids[1]][:2]
        )

        candidate_edge_ids = [e for e in candidate_edge_ids if self.flag[e] == 0]
        return candidate_edge_ids

    def local_two_tensor(self):
        candidate_edge_ids = self.candidate_edge_ids()
        max_v = np.max([self.distance[e] for e in candidate_edge_ids])
        candidate_edge_ids = [
            e for e in candidate_edge_ids if self.distance[e] == max_v
        ]
        edge_id = candidate_edge_ids[0]  # select one

        for i, edge in enumerate(self.psi.edges):
            if edge_id == edge[2]:
                connected_tensor_id = i
            if edge_id in edge[:2]:
                selected_tensor_id = i

        child_tensor_ids = [
            i
            for i, edge in enumerate(self.psi.edges)
            if edge[2] == self.psi.canonical_center_edge_id
        ]

        for child_tensor_id in child_tensor_ids:
            if child_tensor_id != selected_tensor_id:
                not_selected_tensor_id = child_tensor_id

        return (
            edge_id,
            selected_tensor_id,
            connected_tensor_id,
            not_selected_tensor_id,
        )

    def set_flag(self, not_selected_tensor_id):
        if (
            self.flag[self.psi.edges[not_selected_tensor_id][0]] == 1
            and self.flag[self.psi.edges[not_selected_tensor_id][1]] == 1
        ):
            if self.psi.canonical_center_edge_id != self.psi.top_edge_id:
                self.flag[self.psi.canonical_center_edge_id] = 1
        return

    def contract_central_tensors(self):
        central_tensor_ids = self.psi.central_tensor_ids()

        psi1 = tn.Node(self.psi.tensors[central_tensor_ids[0]], backend=self.backend)
        psi2 = tn.Node(self.psi.tensors[central_tensor_ids[1]], backend=self.backend)
        gauge = tn.Node(self.psi.gauge_tensor, backend=self.backend)

        psi1[2] ^ gauge[0]
        gauge[1] ^ psi2[2]

        psi = tn.contractors.auto(
            [psi1, gauge, psi2],
            output_edge_order=[psi1[0], psi1[1], psi2[0], psi2[1]],
        )
        return psi


class TwoSiteUpdater(TwoSiteUpdaterMixin):
    def __init__(self, psi):
        self.psi = psi
        self.backend = "numpy"
        self.flag = self.initial_flag()
        self.distance = self.initial_distance()

    def decompose_two_tensors(
        self,
        psi,
        max_bond_dim,
        opt_structure=0,
        operate_degeneracy=False,
        epsilon=1e-8,
        delta=0.1,
        max_truncation_error=0.0,
        temperature=0.0,
        e=1e-13,
    ):
        edge_order = [0, 1, 2, 3]
        if opt_structure == 0:
            a = psi[0]
            b = psi[1]
            c = psi[2]
            d = psi[3]
            (u, s, v, _) = tn.split_node_full_svd(psi, [a, b], [c, d])
            p = np.diagonal(s.get_tensor())
            ind = np.min([max_bond_dim, len(p)])
            if operate_degeneracy:
                if ind < len(p):
                    while ind > 1:
                        if (np.abs(p[ind] - p[ind - 1]) / (p[ind - 1] + e)) < delta:
                            ind -= 1
                        else:
                            break
        else:
            candidates = [[0, 1, 2, 3], [2, 1, 0, 3], [0, 2, 1, 3]]

            candidate_index = 0
            ps = []
            inds = []
            for edges in candidates:
                psi_ = psi.copy()
                a = psi_[edges[0]]
                b = psi_[edges[1]]
                c = psi_[edges[2]]
                d = psi_[edges[3]]
                (u_, s_, v_, _) = tn.split_node_full_svd(
                    psi_,
                    [a, b],
                    [c, d],
                )
                p_ = np.diagonal(s_.get_tensor())
                ps.append(p_)
                ind = np.min([max_bond_dim, len(p_)])
                if max_truncation_error > 0.0:
                    p_max = p_[0]
                    for i in range(1, ind):
                        if p_[i] / p_max < max_truncation_error:
                            ind = i
                            break
                if operate_degeneracy:
                    if ind < len(p_):
                        while ind > 1:
                            if (
                                np.abs(p_[ind] - p_[ind - 1]) / (p_[ind - 1] + e)
                            ) < delta:
                                ind -= 1
                            else:
                                break
                inds.append(ind)

            ees = [self.entanglement_entropy(probability=p) for p in ps]
            p_truncates = [p[:ind] for p, ind in zip(ps, inds)]
            errors = [1.0 - np.real(np.sum(p_t**2)) for p_t in p_truncates]
            if opt_structure == 1:
                if temperature == 0.0:
                    candidate_index = np.argmin(ees)
                else:
                    weights = np.array(ees) / temperature
                    weights = np.exp(-weights)
                    weights = np.array(weights) / np.sum(weights)
                    candidate_index = np.random.choice(len(ees), p=weights)
            elif opt_structure == 2:
                candidate_index = np.argmin(errors)
                if (
                    np.isclose(errors[candidate_index], errors[0], atol=1e-14)
                    or max_truncation_error > 0.0
                ):
                    candidate_index = np.argmin(ees)

            if np.isclose(ees[candidate_index], ees[0], atol=epsilon):
                candidate_index = 0
            edge_order = candidates[candidate_index]
            p = ps[candidate_index]
            ind = inds[candidate_index]
        # Degeneracy
        a = psi[edge_order[0]]
        b = psi[edge_order[1]]
        c = psi[edge_order[2]]
        d = psi[edge_order[3]]
        (u, s, v, terr) = tn.split_node_full_svd(
            psi, [a, b], [c, d], max_singular_values=ind
        )
        v = v.reorder_edges([v[1], v[2], v[0]])
        u = u.get_tensor()
        v = v.get_tensor()
        s = s.get_tensor()
        p_truncate = np.diagonal(s)
        err = 1.0 - np.real(np.sum(p_truncate**2))
        s = s / np.linalg.norm(s)
        return u, s, v, p, err, edge_order

    def entanglement_entropy_at_physical_bond(self, psi, psi_edges):
        ee_at_physical_bond = {}
        for i, edge in enumerate(psi_edges):
            if edge in self.psi.physical_edges:
                ee_at_physical_bond[edge] = 0.0
        for edge_id in ee_at_physical_bond.keys():
            psi_ = psi.copy()
            matching_index = list({i for i, v in enumerate(psi_edges) if v == edge_id})[
                0
            ]
            non_matching_indices = list(
                {i for i in range(len(psi_edges)) if i != matching_index}
            )
            psi_dag = psi_.copy(conjugate=True)
            for i in non_matching_indices:
                psi_[i] ^ psi_dag[i]
            rho = tn.contractors.auto(
                [psi_, psi_dag],
                output_edge_order=[psi_[matching_index], psi_dag[matching_index]],
            )
            (u, s, v, terr) = tn.split_node_full_svd(rho, [rho[0]], [rho[1]])
            p_ = np.diagonal(s.get_tensor())
            p_ = np.sqrt(p_)
            ee_at_physical_bond[edge_id] = self.entanglement_entropy(probability=p_)
        return ee_at_physical_bond

    def set_ttn_properties_at_one_tensor(self, edge_id, selected_tensor_id):
        # update_ttn_properties
        self.psi.canonical_center_edge_id = edge_id
        out_selected_inds = []
        for i, e in enumerate(self.psi.edges[selected_tensor_id]):
            if e == edge_id:
                canonical_center_ind = i
            else:
                out_selected_inds.append(i)
        self.psi.tensors[selected_tensor_id] = self.psi.tensors[
            selected_tensor_id
        ].transpose(
            out_selected_inds + [canonical_center_ind],
        )
        self.psi.edges[selected_tensor_id] = [
            self.psi.edges[selected_tensor_id][i] for i in out_selected_inds
        ] + [edge_id]
        for i, e in enumerate(self.psi.edges[selected_tensor_id]):
            self.psi.edge_dims[e] = self.psi.tensors[selected_tensor_id].shape[i]
        return



from typing import List, Union
import numpy as np


class Observable:
    """A class for observables."""

    def __init__(
        self,
        indices: List[int],
        operators_list: List[List[str]],
        coef_list: Union[List[float], List[complex]],
    ):
        """Initialize an Observable object.

        Args:
            indices : Indices of the observable.
            operators_list : List of operators, "S+", "S-", "Sz", "Sx" and "Sy".
            coef_list : Coefficients for each operator.
        """
        self.indices = indices
        self.operators_list = operators_list
        self.coef_list = coef_list

        self.indices_num = len(indices)
        self.operators_num = len(operators_list)


def spin_ind(spin_num):
    if isinstance(spin_num, str):
        if spin_num.startswith("S="):
            spin_value_str = spin_num[2:]
        else:
            print("=" * 50)
            print("⚠️  Error: Invalid spin string format. Expected format 'S=...'.")
            print("=" * 50)
            exit()

        if "/" in spin_value_str:
            numerator, denominator = spin_value_str.split("/")
            spin_value = int(numerator) / int(denominator)
        else:
            spin_value = float(spin_value_str)
    elif isinstance(spin_num, (int, float)):
        spin_value = spin_num
    else:
        print("=" * 50)
        print("⚠️  Error: Invalid type for spin_num. Expected a string or a number.")
        print("=" * 50)
        exit()

    return spin_value


def bare_spin_operator(spin, spin_num):
    spin_value = spin_ind(spin_num)  # Get both string form and numeric value
    # Check if spin_value is valid
    if not (spin_value == int(spin_value) or spin_value == int(spin_value) + 0.5):
        print("=" * 50)
        print(
            "⚠️  Error: Invalid spin number. Spin number must be an integer or half-integer."
        )
        print("=" * 50)
        exit()
    dim = int(2 * spin_value + 1)  # Dimension of the matrix based on the spin value
    if spin == "S+":
        # Construct the raising operator S+
        S_plus = np.zeros((dim, dim), dtype=np.complex128)
        for m in range(dim - 1):
            S_plus[m, m + 1] = np.sqrt(
                spin_value * (spin_value + 1) - (spin_value - m) * (spin_value - m - 1)
            )
        return S_plus

    elif spin == "S-":
        # Construct the raising operator S+
        S_plus = np.zeros((dim, dim), dtype=np.complex128)
        for m in range(dim - 1):
            S_plus[m, m + 1] = np.sqrt(
                spin_value * (spin_value + 1) - (spin_value - m) * (spin_value - m - 1)
            )
        S_minus = np.transpose(S_plus)
        return S_minus

    elif spin == "Sz":
        # Construct the Sz operator
        Sz = np.zeros((dim, dim), dtype=np.complex128)
        for m in range(dim):
            Sz[m, m] = spin_value - m
        return Sz

    elif spin == "Sx":
        Sp = bare_spin_operator("S+", spin_num)
        Sm = bare_spin_operator("S-", spin_num)
        Sx = (Sp + Sm) / 2.0
        return Sx

    elif spin == "Sy":
        Sp = bare_spin_operator("S+", spin_num)
        Sm = bare_spin_operator("S-", spin_num)
        Sy = (Sp - Sm) / 2.0j
        return Sy


def spin_dof(spin_num):
    spin_value = spin_ind(spin_num)
    dim = int(2 * spin_value + 1)
    return dim



from copy import deepcopy
from typing import Dict, Tuple

import numpy as np
import tensornetwork as tn

from ttnopt.src.Hamiltonian import Hamiltonian
from ttnopt.src.PhysicsEngine import PhysicsEngine
from ttnopt.src.TTN import TreeTensorNetwork


class GroundStateSearch(PhysicsEngine):
    """A class for ground state search algorithm based on DMRG.
    Args:
        psi (TreeTensorNetwork): The quantum state.
        hamiltonians (Hamiltonian): Hamiltonian which is list of Observable.
        init_bond_dim (int, optional): Initial bond dimension. Defaults to 4.
        max_bond_dim (int, optional): Maximum bond dimension. Defaults to 16.
    """

    def __init__(
        self,
        psi: TreeTensorNetwork,
        hamiltonian: Hamiltonian,
        init_bond_dim: int = 4,
        max_bond_dim: int = 16,
        energy_degeneracy_threshold: float = 1e-13,
        entanglement_degeneracy_threshold: float = 0.1,
    ):
        """Initialize a DMRG object.

        Args:
            psi : The quantum state.
            hamiltonians : The Hamiltonian.
            init_bond_dim : Initial bond dimension.
            max_bond_dim : Maximum bond dimension.
            truncation_error : Maximum truncation error.
        """
        self.energy: Dict[int, float] = {}
        self.entanglement: Dict[int, float] = {}
        self.error: Dict[int, float] = {}
        self.one_site_expval: Dict[int, Dict[str, float]] = {}
        self.two_site_expval: Dict[Tuple[int, int], Dict[str, float]] = {}

        super().__init__(
            psi,
            hamiltonian,
            init_bond_dim,
            max_bond_dim,
            energy_degeneracy_threshold,
            entanglement_degeneracy_threshold,
        )

    def run(
        self,
        opt_structure: int = 0,
        energy_convergence_threshold: float = 1e-8,
        entanglement_convergence_threshold: float = 1e-8,
        max_num_sweep: int = 10,
        converged_count: int = 2,
        eval_onesite_expval: bool = False,
        eval_twosite_expval: bool = False,
        temperature: float = 0.0,
        tau: int = 0,
    ):
        """Run DMRG algorithm.

        Args:
            opt_structure (bool, optional): If optimize the tree structure or not. Defaults to False.
            energy_convergence_threshold (float, optional): Energy threshold for convergence. Defaults to 1e-8.
            entanglement_convergence_threshold (float, optional): Entanglement entropy threshold for automatic optimization. Defaults to 1e-8.
            converged_count (int, optional): Converged count. Defaults to 1.
            eval_onesite_expval (bool): If evaluate one-site expectation value or not.
            eval_twosite_expval (bool): If evaluate two-site expectation value or not.
        """
        energy_at_edge: Dict[int, float] = {}
        _energy_at_edge: Dict[int, float] = {}
        ee_at_edge: Dict[int, float] = {}
        _ee_at_edge: Dict[int, float] = {}
        _error_at_edge: Dict[int, float] = {}
        onesite_expval: Dict[int, Dict[str, float]] = {}
        twosite_expval: Dict[Tuple[int, int], Dict[str, float]] = {}

        edges, _edges = deepcopy(self.psi.edges), deepcopy(self.psi.edges)

        converged_num = 0

        if tau == 0:
            tau = max_num_sweep // 2 + 1
        for sweep_num in range(max_num_sweep):
            temp = temperature * (2 ** (-sweep_num / tau))
            if converged_num > converged_count:
                break

            energy_at_edge = deepcopy(_energy_at_edge)
            ee_at_edge = deepcopy(_ee_at_edge)
            edges = deepcopy(_edges)

            self.distance = self.initial_distance()
            self.flag = self.initial_flag()

            (
                _edge_id,
                _selected_tensor_id,
                _connected_tensor_id,
                _not_selected_tensor_id,
            ) = self.local_two_tensor()

            print("Sweep count: " + str(sweep_num + 1))
            while True:
                edge_id = _edge_id
                selected_tensor_id = _selected_tensor_id
                connected_tensor_id = _connected_tensor_id
                not_selected_tensor_id = _not_selected_tensor_id
                # absorb gauge tensor
                iso = tn.Node(self.psi.tensors[selected_tensor_id])
                gauge = tn.Node(self.psi.gauge_tensor)
                iso[2] ^ gauge[0]
                iso = tn.contractors.auto(
                    [iso, gauge], output_edge_order=[iso[0], iso[1], gauge[1]]
                )
                self.psi.tensors[selected_tensor_id] = iso.get_tensor()

                self.set_ttn_properties_at_one_tensor(edge_id, selected_tensor_id)

                self._set_edge_spin(not_selected_tensor_id)

                self._set_block_hamiltonian(not_selected_tensor_id)

                ground_state_order = [selected_tensor_id, connected_tensor_id]
                ground_state, energy = self.lanczos(ground_state_order)
                psi_edges = (
                    self.psi.edges[selected_tensor_id][:2]
                    + self.psi.edges[connected_tensor_id][:2]
                )

                u, s, v, probability, error, edge_order = self.decompose_two_tensors(
                    ground_state,
                    self.max_bond_dim,
                    opt_structure=opt_structure,
                    temperature=temp,
                    operate_degeneracy=True,
                    epsilon=entanglement_convergence_threshold,
                    delta=self.entanglement_degeneracy_threshold,
                )

                self.psi.tensors[selected_tensor_id] = u
                self.psi.tensors[connected_tensor_id] = v
                self.psi.gauge_tensor = s
                (
                    self.psi.edges[selected_tensor_id][0],
                    self.psi.edges[selected_tensor_id][1],
                ) = (
                    psi_edges[edge_order[0]],
                    psi_edges[edge_order[1]],
                )
                (
                    self.psi.edges[connected_tensor_id][0],
                    self.psi.edges[connected_tensor_id][1],
                ) = (
                    psi_edges[edge_order[2]],
                    psi_edges[edge_order[3]],
                )

                self.distance = self.initial_distance()
                _energy_at_edge[self.psi.canonical_center_edge_id] = energy
                print(energy)
                ee = self.entanglement_entropy(probability)
                _ee_at_edge[self.psi.canonical_center_edge_id] = ee
                ee_dict = self.entanglement_entropy_at_physical_bond(
                    ground_state, psi_edges
                )
                for key in ee_dict.keys():
                    _ee_at_edge[key] = ee_dict[key]
                _error_at_edge[self.psi.canonical_center_edge_id] = error

                if self.candidate_edge_ids() == []:
                    break

                (
                    _edge_id,
                    _selected_tensor_id,
                    _connected_tensor_id,
                    _not_selected_tensor_id,
                ) = self.local_two_tensor()
                self.set_flag(_not_selected_tensor_id)
                # eval expval
                if self.flag[_not_selected_tensor_id]:
                    if eval_onesite_expval:
                        onesite_expval_dict = self.expval_onesite(
                            _not_selected_tensor_id,
                            ground_state,
                            ground_state_order,
                        )
                        for key in onesite_expval_dict.keys():
                            onesite_expval[key] = onesite_expval_dict[key]
                    if eval_twosite_expval:
                        twosite_expval_dict = self.expval_twosite(
                            _not_selected_tensor_id,
                            ground_state,
                            ground_state_order,
                        )
                        for key in twosite_expval_dict.keys():
                            twosite_expval[key] = twosite_expval_dict[key]

            if eval_onesite_expval:
                for i in self.psi.central_tensor_ids():
                    onesite_expval_dict = self.expval_onesite(
                        i, ground_state, ground_state_order
                    )
                    for key in onesite_expval_dict.keys():
                        onesite_expval[key] = onesite_expval_dict[key]
            if eval_twosite_expval:
                for i in self.psi.central_tensor_ids():
                    twosite_expval_dict = self.expval_twosite(
                        i, ground_state, ground_state_order
                    )
                    for key in twosite_expval_dict.keys():
                        twosite_expval[key] = twosite_expval_dict[key]
                twosite_expval_dict = self.expval_twosite_origin(
                    twosite_expval.keys(), ground_state, ground_state_order
                )
                for key in twosite_expval_dict.keys():
                    twosite_expval[key] = twosite_expval_dict[key]

            _edges = deepcopy(self.psi.edges)

            sweep_num += 1
            if sweep_num > 2:
                diff_energy = [
                    np.abs(1 - _energy_at_edge[key] / energy_at_edge[key])
                    for key in energy_at_edge.keys()
                ]
                diff_ee = [
                    np.abs(ee_at_edge[key] - _ee_at_edge[key])
                    for key in ee_at_edge.keys()
                ]
                if all(
                    [
                        set(edge[:2]) == set(_edge[:2]) and edge[2] == _edge[2]
                        for edge, _edge in zip(edges, _edges)
                    ]
                ):
                    if all(
                        [
                            energy < energy_convergence_threshold
                            for energy in diff_energy
                        ]
                    ) and all(
                        [ee < entanglement_convergence_threshold for ee in diff_ee]
                    ):
                        converged_num += 1
        print("Converged")

        self.energy = _energy_at_edge
        self.entanglement = _ee_at_edge
        self.error = _error_at_edge
        self.one_site_expval = onesite_expval
        self.two_site_expval = twosite_expval
        return 0



import numpy as np
import tensornetwork as tn


def init_tensors_mps(state, max_bond, truncate, min_bond=2):  # TODO
    """_summary_

    Args:
        state tensor: initial_state where the shape is (bond_dim[0], bond_dim[1], ..., bond_dim[N])
        size (_type_): _description_
    Returns:
        tensors: List[tn.Node]: List of tensors
    """
    tensors = []
    size = len(state.shape)
    for i in range(size):
        if i == 0:
            print("0")
        elif i == size - 1:
            tensor = tn.Node(state[i], backend="numpy")
        else:
            tensor = tn.Node(state[i], backend="numpy")
        tensors.append(tensor)
    return tensors


def get_upper_bond_dim(lower_bond_dims, max_bond_dim):
    bond_dim = np.prod(lower_bond_dims)
    if bond_dim > max_bond_dim:
        bond_dim = max_bond_dim
    return bond_dim



import numpy as np
import tensornetwork as tn
from tensornetwork import U1Charge, BlockSparseTensor

from ttnopt.src.TwoSiteUpdater import TwoSiteUpdaterMixin


class TwoSiteUpdaterSparse(TwoSiteUpdaterMixin):
    def __init__(self, psi):
        self.psi = psi
        self.backend = "symmetric"
        self.flag = self.initial_flag()
        self.distance = self.initial_distance()

    def decompose_two_tensors(
        self,
        psi,
        max_bond_dim,
        opt_structure=0,
        epsilon=1e-8,
        delta=0.1,
        temperature=0.0,
    ):
        edge_order = [0, 1, 2, 3]
        psi_last = psi.copy()
        if not opt_structure:
            a = psi[0]
            b = psi[1]
            c = psi[2]
            d = psi[3]
            e = psi[4]
            (u, s, v, _) = tn.split_node_full_svd(psi, [e, a, b], [c, d])
            p = s.tensor.data
            p = np.sort(p)[::-1]
            p = p[p > 0.0]
            ind = np.min([max_bond_dim, len(p)])
            if ind < len(p):
                while ind > 1:
                    if (np.abs(p[ind] - p[ind - 1]) / (p[ind - 1] + e)) < delta:
                        ind -= 1
                    else:
                        break
        else:
            candidates = [[0, 1, 2, 3], [2, 1, 0, 3], [0, 2, 1, 3]]
            candidate_index = 0
            ps = []
            inds = []
            for edges in candidates:
                psi_ = psi.copy()
                a = psi_[edges[0]]
                b = psi_[edges[1]]
                c = psi_[edges[2]]
                d = psi_[edges[3]]
                e = psi_[4]
                (u_, s_, v_, terr) = tn.split_node_full_svd(psi_, [e, a, b], [c, d])

                p_ = s_.tensor.data
                p_ = np.sort(p_)[::-1]
                p_ = p_[p_ > 0.0]
                ps.append(p_)
                # diagonal
                ind = np.min([max_bond_dim, len(p_)])
                if ind < len(p_):
                    while ind > 1:
                        if (np.abs(p_[ind] - p_[ind - 1]) / p_[ind]) < delta:
                            ind -= 1
                        else:
                            break
                inds.append(ind)

            ees = [self.entanglement_entropy(probability=p) for p in ps]
            p_truncates = [p[:max_bond_dim] for p in ps]
            errors = [1.0 - np.real(np.sum(p_t**2)) for p_t in p_truncates]
            if opt_structure == 1:
                if temperature == 0.0:
                    candidate_index = np.argmin(ees)
                else:
                    weights = np.array(ees) / temperature
                    weights = np.exp(-weights)
                    weights = np.array(weights) / np.sum(weights)
                    candidate_index = np.random.choice(len(ees), p=weights)
            elif opt_structure == 2:
                candidate_index = np.argmin(errors)
                if np.isclose(errors[candidate_index], errors[0], atol=1e-14):
                    candidate_index = np.argmin(ees)

            if np.isclose(ees[candidate_index], ees[0], atol=epsilon):
                candidate_index = 0
            edge_order = candidates[candidate_index]
            p = ps[candidate_index]
            ind = inds[candidate_index]

        a = psi_last[edge_order[0]]
        b = psi_last[edge_order[1]]
        c = psi_last[edge_order[2]]
        d = psi_last[edge_order[3]]
        e = psi_last[4]
        (u, s, v, terr) = tn.split_node_full_svd(
            psi_last, [e, a, b], [c, d], max_singular_values=ind
        )
        u = u.reorder_edges([u[1], u[2], u[3], u[0]])
        v = v.reorder_edges([v[1], v[2], v[0]])

        u_tensor = u.tensor
        u_tensor.contiguous(inplace=True)
        s_tensor = s.tensor / np.linalg.norm(s.tensor.data)
        s_tensor.contiguous(inplace=True)
        v_tensor = v.tensor
        v_tensor.contiguous(inplace=True)
        p_truncate = s.tensor.data

        s = tn.Node(s_tensor, backend=self.backend)
        u = tn.Node(u_tensor, backend=self.backend)
        a, ss, b, terr = tn.split_node_full_svd(
            u,
            [
                u[0],
                u[1],
            ],
            [u[2], u[3]],
        )
        u_tensor = a.tensor
        u_tensor.contiguous(inplace=True)
        u_tensor = BlockSparseTensor(
            data=u_tensor.data,
            charges=[
                U1Charge(u_tensor.flat_charges[0].charges.flatten()),
                U1Charge(u_tensor.flat_charges[1].charges.flatten()),
                U1Charge(-u_tensor.flat_charges[2].charges.flatten()),
            ],
            flows=[True, True, False],
        )
        s[0] ^ b[1]
        s = tn.contractors.auto([s, b], output_edge_order=[b[0], s[1], b[2]])
        s_tensor = s.tensor
        s_tensor.contiguous(inplace=True)
        s_tensor = BlockSparseTensor(
            data=s_tensor.data,
            charges=[
                U1Charge(-s_tensor.flat_charges[0].charges.flatten()),
                U1Charge(s_tensor.flat_charges[1].charges.flatten()),
                U1Charge(s_tensor.flat_charges[2].charges.flatten()),
            ],
            flows=[True, True, False],
        )
        err = 1.0 - np.real(np.sum(p_truncate**2))
        return (
            u_tensor,
            s_tensor,
            v_tensor,
            p,
            err,
            edge_order,
        )

    def entanglement_entropy_at_physical_bond(self, psi, psi_edges):
        ee_at_physical_bond = {}
        for i, edge in enumerate(psi_edges):
            if edge in self.psi.physical_edges:
                ee_at_physical_bond[edge] = 0.0
        for edge_id in ee_at_physical_bond.keys():
            psi_ = psi.copy()
            matching_index = list({i for i, v in enumerate(psi_edges) if v == edge_id})[
                0
            ]
            non_matching_indices = list(
                {i for i in range(len(psi_edges)) if i != matching_index}
            )
            psi_dag = psi_.copy(conjugate=True)
            for i in non_matching_indices:
                psi_[i] ^ psi_dag[i]

            psi_[4] ^ psi_dag[4]

            rho = tn.contractors.auto(
                [psi_, psi_dag],
                output_edge_order=[psi_[matching_index], psi_dag[matching_index]],
            )
            (u, s, v, terr) = tn.split_node_full_svd(rho, [rho[0]], [rho[1]])
            p_ = s.tensor.data
            p_ = p_[p_ > 0.0]
            p_ = np.sqrt(p_)
            ee_at_physical_bond[edge_id] = self.entanglement_entropy(p_)
        return ee_at_physical_bond

    def set_ttn_properties_at_one_tensor(self, edge_id, selected_tensor_id):
        # update_ttn_properties
        self.psi.canonical_center_edge_id = edge_id
        out_selected_inds = []
        for i, e in enumerate(self.psi.edges[selected_tensor_id]):
            if e == edge_id:
                canonical_center_ind = i
            else:
                out_selected_inds.append(i)
        self.psi.tensors[selected_tensor_id] = self.psi.tensors[
            selected_tensor_id
        ].transpose(
            out_selected_inds + [canonical_center_ind] + [3],
        )
        self.psi.edges[selected_tensor_id] = [
            self.psi.edges[selected_tensor_id][i] for i in out_selected_inds
        ] + [edge_id]
        for i, e in enumerate(self.psi.edges[selected_tensor_id]):
            self.psi.edge_dims[e] = self.psi.tensors[selected_tensor_id].shape[i]
        return



import numpy as np
import tensornetwork as tn


def get_renormalization_sequence(edges, top_edge_id):
    tensor_flag = [0 for _ in range(len(edges))]
    top_child_ids = [i for i in range(len(edges)) if edges[i][2] == top_edge_id]

    stack = [(-1, top_child_ids)]
    sequence = []
    while stack:
        parent_id, child_ids = stack[-1]
        if all(tensor_flag[child_id] == 1 for child_id in child_ids) or not child_ids:
            if parent_id != -1:
                sequence.append(parent_id)
                tensor_flag[parent_id] = 1
            stack.pop()
        else:
            for child_id in child_ids:
                if tensor_flag[child_id] == 0:
                    stack.append((child_id, get_child_tensor_id(child_id, edges)))
    return sequence


def get_child_tensor_id(tensor_id, edges):
    child_tensor_id = []
    for i, edge in enumerate(edges):
        if edge[2] in edges[tensor_id][:2]:
            child_tensor_id.append(i)
    return child_tensor_id


def get_bare_edges(edge_id, edges, physical_edges):
    bare_spins = []
    if edge_id in physical_edges:
        return [edge_id]
    else:
        child_ids = [i for i, edge in enumerate(edges) if edge_id == edge[2]]
        while len(child_ids) > 0:
            new_child = []
            for child_id in child_ids:
                for edge in edges[child_id][:2]:
                    if edge in physical_edges:
                        bare_spins.append(edge)
                new_child += get_child_tensor_id(child_id, edges)
            child_ids = new_child
    return bare_spins


def inner_product(u, v):
    u = u.copy(conjugate=True)
    u[0] ^ v[0]
    u[1] ^ v[1]
    u[2] ^ v[2]
    u[3] ^ v[3]
    return tn.contractors.auto([u, v]).get_tensor()


def inner_product_sparse(u, v):
    u = u.copy(conjugate=True)
    u[0] ^ v[0]
    u[1] ^ v[1]
    u[2] ^ v[2]
    u[3] ^ v[3]
    u[4] ^ v[4]
    prod = tn.contractors.auto([u, v]).tensor.todense()
    return np.real(prod)[0]



import math
import numpy as np

from typing import Optional, List
from ttnopt.src.Observable import Observable


class Hamiltonian:
    """A class for Hamiltonian.
    This class is used to store Hamiltonian as a List of Observable.
    """

    def __init__(
        self,
        system_size: int,
        spin_size: List[str],
        model: str,
        interaction_indices: List[List[int]],
        interaction_coefs: List[List[float]],
        magnetic_field_X_indices: Optional[List[int]] = None,
        magnetic_field_X: Optional[List[float]] = None,
        magnetic_field_Y_indices: Optional[List[int]] = None,
        magnetic_field_Y: Optional[List[float]] = None,
        magnetic_field_Z_indices: Optional[List[int]] = None,
        magnetic_field_Z: Optional[List[float]] = None,
        ion_anisotropy_indices: Optional[List[int]] = None,
        ion_anisotropy: Optional[List[float]] = None,
        dzyaloshinskii_moriya_X_indices: Optional[List[List[int]]] = None,
        dzyaloshinskii_moriya_X: Optional[List[float]] = None,
        dzyaloshinskii_moriya_Y_indices: Optional[List[List[int]]] = None,
        dzyaloshinskii_moriya_Y: Optional[List[float]] = None,
        dzyaloshinskii_moriya_Z_indices: Optional[List[List[int]]] = None,
        dzyaloshinskii_moriya_Z: Optional[List[float]] = None,
        sod_X_indices: Optional[List[List[int]]] = None,
        sod_X: Optional[List[float]] = None,
        sod_Y_indices: Optional[List[List[int]]] = None,
        sod_Y: Optional[List[float]] = None,
        sod_Z_indices: Optional[List[List[int]]] = None,
        sod_Z: Optional[List[float]] = None,
    ):
        """
        Args:
            system_size (int): The size of the system, indicating the number of sites or particles.
            spin_size (List[str]): A list of spin types or values, each corresponding to the spin state of each site in the system.
            model (str): The type of Hamiltonian model, which determines the specific interactions and parameters used (e.g., "Heisenberg", "Ising").
            interaction_indices (List[List[int]]): A nested list of integer pairs, where each sub-list specifies two indices corresponding
                to sites or particles between which an interaction occurs.
            interaction_coefs (List[List[float]]): A nested list where each sub-list contains the coefficients for the interactions
                specified in `interaction_indices`, defining the strength and nature of each interaction.
            magnetic_field_indices (Optional[List[int]], optional): A list of site indices at which the magnetic field is applied.
                If None, no magnetic field is applied to any specific site. Defaults to None.
            magnetic_field (Optional[List[float]], optional): A list of magnetic field strengths, where each value corresponds
                to a specific site defined in `magnetic_field_indices`. Defaults to None.
            magnetic_field_axis (Optional[str], optional): The axis along which the magnetic field is applied, specified as a string
                ("x", "y", or "z"). Defaults to None.
            ion_anisotropy_indices (Optional[List[int]], optional): A list of indices specifying the sites with single-ion anisotropy.
                Defaults to None.
            ion_anisotropy (Optional[List[float]], optional): A list of single-ion anisotropy values corresponding to each site specified
                in `ion_anisotropy_indices`. Defaults to None.
            dzyaloshinskii_moriya_indices (Optional[List[List[int]]], optional): A nested list of integer pairs, where each sub-list specifies
                two indices corresponding to sites with Dzyaloshinskii-Moriya (DM) interaction. Defaults to None.
            dzyaloshinskii_moriya (Optional[List[List[float]]], optional): A list of DM interaction strengths, where each value corresponds
                to the pair specified in `dzyaloshinskii_moriya_indices`. Defaults to None.
            sod_indices (Optional[List[List[int]]], optional): A nested list of integer pairs, where each sub-list specifies two indices
                two indices corresponding to sites with second-order symmetric off-diagonal anisotropy interaction. Defaults to None.
            sod (Optional[List[List[float]]], optional): A list of SOD interaction strengths, where each value corresponds
                to the pair specified in `sod_indices`. Defaults to None.
        """

        self.system_size = system_size
        self.spin_size = {i: spin_size[i] for i in range(self.system_size)}
        self.observables = []

        if model == "XXZ":
            for i, coef in zip(interaction_indices, interaction_coefs):
                if all([c == 0.0 for c in coef]):
                    continue
                operator_list = []
                coef_list = []
                if not coef[0] == 0.0:
                    operator_list.append(["S+", "S-"])
                    coef_list.append(coef[0] / 2.0)
                    operator_list.append(["S-", "S+"])
                    coef_list.append(coef[0] / 2.0)
                if not coef[0] == 0.0 and not coef[1] == 0.0:
                    operator_list.append(["Sz", "Sz"])
                    coef_list.append(coef[0] * coef[1])
                ob = Observable(i, operator_list, coef_list)
                self.observables.append(ob)
        if model == "XYZ":
            for i, coef in zip(interaction_indices, interaction_coefs):
                if all([c == 0.0 for c in coef]):
                    continue
                operator_list = []
                coef_list = []
                if not coef[0] == 0.0:
                    operator_list.append(["Sx", "Sx"])
                    coef_list.append(coef[0])
                if not coef[1] == 0.0:
                    operator_list.append(["Sy", "Sy"])
                    coef_list.append(coef[1])
                if not coef[2] == 0.0:
                    operator_list.append(["Sz", "Sz"])
                    coef_list.append(coef[2])
                ob = Observable(i, operator_list, coef_list)
                self.observables.append(ob)

        if magnetic_field_X is not None and magnetic_field_X_indices is not None:
            for idx, c in zip(magnetic_field_X_indices, magnetic_field_X):
                if not c == 0.0:
                    ob = Observable([idx], [["Sx"]], [-c])
                    self.observables.append(ob)

        if magnetic_field_Y is not None and magnetic_field_Y_indices is not None:
            for idx, c in zip(magnetic_field_Y_indices, magnetic_field_Y):
                if not c == 0.0:
                    ob = Observable([idx], [["Sy"]], [-c])
                    self.observables.append(ob)

        if magnetic_field_Z is not None and magnetic_field_Z_indices is not None:
            for idx, c in zip(magnetic_field_Z_indices, magnetic_field_Z):
                if not c == 0.0:
                    ob = Observable([idx], [["Sz"]], [-c])
                    self.observables.append(ob)

        if ion_anisotropy is not None and ion_anisotropy_indices is not None:
            for idx, c in zip(ion_anisotropy_indices, ion_anisotropy):
                if not c == 0.0:
                    ob = Observable([idx], [["Sz2"]], [c])
                    self.observables.append(ob)

        if (
            dzyaloshinskii_moriya_X is not None
            and dzyaloshinskii_moriya_X_indices is not None
        ):
            for i, c in zip(dzyaloshinskii_moriya_X_indices, dzyaloshinskii_moriya_X):
                if not c == 0.0:
                    ob = Observable(i, [["Sy", "Sz"], ["Sz", "Sy"]], [c, -c])
                    self.observables.append(ob)

        if (
            dzyaloshinskii_moriya_Y is not None
            and dzyaloshinskii_moriya_Y_indices is not None
        ):
            for i, c in zip(dzyaloshinskii_moriya_Y_indices, dzyaloshinskii_moriya_Y):
                if not c == 0.0:
                    ob = Observable(i, [["Sz", "Sx"], ["Sx", "Sz"]], [c, -c])
                    self.observables.append(ob)

        if (
            dzyaloshinskii_moriya_Z is not None
            and dzyaloshinskii_moriya_Z_indices is not None
        ):
            for i, c in zip(dzyaloshinskii_moriya_Z_indices, dzyaloshinskii_moriya_Z):
                if not c == 0.0:
                    cc = np.array([c / 2.0j])
                    ob = Observable(
                        i,
                        [["S-", "S+"], ["S+", "S-"]],
                        [cc.item(), -cc.item()],
                    )
                    self.observables.append(ob)

        if sod_X is not None and sod_X_indices is not None:
            for i, c in zip(sod_X_indices, sod_X):
                if not c == 0.0:
                    ob = Observable(
                        i,
                        [["Sy", "Sz"], ["Sz", "Sy"]],
                        [c, c],
                    )
                    self.observables.append(ob)

        if sod_Y is not None and sod_Y_indices is not None:
            for i, c in zip(sod_Y_indices, sod_Y):
                if not c == 0.0:
                    ob = Observable(
                        i,
                        [["Sz", "Sx"], ["Sx", "Sz"]],
                        [c, c],
                    )
                    self.observables.append(ob)

        if sod_Z is not None and sod_Z_indices is not None:
            for i, c in zip(sod_Z_indices, sod_Z):
                if not c == 0.0:
                    ob = Observable(
                        i,
                        [["Sx", "Sy"], ["Sy", "Sz"]],
                        [c, c],
                    )
                    self.observables.append(ob)



# ground_state_search.py
import argparse
import itertools
import os
from pathlib import Path

import numpy as np
import pandas as pd
import yaml
from dotmap import DotMap

from ttnopt.hamiltonian import hamiltonian
from ttnopt.src import GroundStateSearch, GroundStateSearchSparse, TreeTensorNetwork


def ground_state_search():
    parser = argparse.ArgumentParser(description="Ground state search simulation")
    parser.add_argument(
        "config_file", type=str, help="Path to the YAML configuration file"
    )
    args = parser.parse_args()

    with open(args.config_file, "r") as file:
        config = yaml.safe_load(file)
    config = DotMap(config)

    psi = TreeTensorNetwork.mps(config.system.N)
    if config.numerics.init_tree == 1:
        if config.system.N > 0 and (config.system.N & (config.system.N - 1)) == 0:
            psi = TreeTensorNetwork.tree(config.system.N)
        else:
            print("=" * 50)
            print("⚠️  Note: N is not a power of 2.")
            print("     Using MPS structure as the default.")
            print("=" * 50)

    ham = hamiltonian(config.system)

    numerics = config.numerics

    path = (
        Path(config.output.dir)
        if not isinstance(config.output.dir, DotMap)
        else Path("./")
    )
    os.makedirs(path, exist_ok=True)

    save_onesite_expval = (
        config.output.single_site
        if not isinstance(config.output.single_site, DotMap)
        else False
    )
    save_twosite_expval = (
        config.output.two_site
        if not isinstance(config.output.two_site, DotMap)
        else False
    )

    u1_symmetry = True if not isinstance(numerics.U1_symmetry, DotMap) else False

    if u1_symmetry and config.model.type == "XYZ":
        print("=" * 50)
        print("⚠️  Error: U1 symmetry is not supported for the XYZ model.")
        print("     Please set the U1 symmetry to False or change the model to XXZ.")
        print("=" * 50)
        exit()
    if u1_symmetry and not (isinstance(config.system.MF_X, DotMap)):
        print("=" * 50)
        print("⚠️  Error: U1 symmetry is not allowed for the X magnetic field.")
        print("=" * 50)
        exit()
    if u1_symmetry and not (isinstance(config.system.MF_Y, DotMap)):
        print("=" * 50)
        print("⚠️  Error: U1 symmetry is not allowed for the Y magnetic field.")
        print("=" * 50)
        exit()
    if u1_symmetry and not isinstance(config.system.DM_X, DotMap):
        print("=" * 50)
        print(
            "⚠️  Error: U1 symmetry is not allowed for the X term of Dzyaloshinskii-Moriya interaction."
        )
        print("=" * 50)
        exit()
    if u1_symmetry and not isinstance(config.system.DM_Y, DotMap):
        print("=" * 50)
        print(
            "⚠️  Error: U1 symmetry is not allowed for the Y term of Dzyaloshinskii-Moriya interaction."
        )
        print("=" * 50)
        exit()
    if u1_symmetry and not isinstance(config.system.SOD_X, DotMap):
        print("=" * 50)
        print(
            "⚠️  Error: U1 symmetry is not allowed for the X term of symmetric exchange anisotropy interaction."
        )
        print("=" * 50)
        exit()
    if u1_symmetry and not isinstance(config.system.SOD_Y, DotMap):
        print("=" * 50)
        print(
            "⚠️  Error: U1 symmetry is not allowed for the Y term of symmetric exchange anisotropy interaction."
        )
        print("=" * 50)
        exit()
    if u1_symmetry and not isinstance(config.system.SOD_Z, DotMap):
        print("=" * 50)
        print(
            "⚠️  Error: U1 symmetry is not allowed for the Z term of symmetric exchange anisotropy interaction."
        )
        print("=" * 50)
        exit()

    opt_structure = (
        numerics.opt_structure.type
        if (isinstance(numerics.opt_structure.type, int))
        else 0
    )

    temperature = (
        float(numerics.opt_structure.temperature)
        if (
            isinstance(numerics.opt_structure.temperature, float)
            or isinstance(numerics.opt_structure.temperature, int)
        )
        else 0.0
    )

    seed = (
        numerics.opt_structure.seed
        if isinstance(numerics.opt_structure.seed, int)
        else 0
    )
    if isinstance(numerics.energy_degeneracy_threshold, DotMap):
        energy_degeneracy_threshold = 1.0e-8
    else:
        energy_degeneracy_threshold = float(numerics.energy_degeneracy_threshold)
    if isinstance(numerics.entanglement_degeneracy_threshold, DotMap):
        entanglement_degeneracy_threshold = 1.0e-8
    else:
        entanglement_degeneracy_threshold = float(
            numerics.entanglement_degeneracy_threshold
        )
    if isinstance(numerics.energy_convergence_threshold, DotMap):
        energy_convergence_threshold = 1.0e-8
    else:
        energy_convergence_threshold = float(numerics.energy_convergence_threshold)
    if isinstance(numerics.entanglement_convergence_threshold, DotMap):
        entanglement_convergence_threshold = 1.0e-8
    else:
        entanglement_convergence_threshold = float(
            numerics.entanglement_convergence_threshold
        )

    np.random.seed(seed)

    tau = (
        numerics.opt_structure.tau
        if isinstance(numerics.opt_structure.tau, int)
        else numerics.max_bond_dimensions[0] // 2
    )

    if isinstance(numerics.energy_degeneracy_threshold, DotMap):
        energy_degeneracy_threshold = 1.0e-8
    else:
        energy_degeneracy_threshold = float(numerics.energy_degeneracy_threshold)
    if isinstance(numerics.entanglement_degeneracy_threshold, DotMap):
        entanglement_degeneracy_threshold = 1.0e-8
    else:
        entanglement_degeneracy_threshold = float(
            numerics.entanglement_degeneracy_threshold
        )
    if isinstance(numerics.energy_convergence_threshold, DotMap):
        energy_convergence_threshold = 1.0e-8
    else:
        energy_convergence_threshold = float(numerics.energy_convergence_threshold)
    if isinstance(numerics.entanglement_convergence_threshold, DotMap):
        entanglement_convergence_threshold = 1.0e-8
    else:
        entanglement_convergence_threshold = float(
            numerics.entanglement_convergence_threshold
        )

    if u1_symmetry:
        gss = GroundStateSearchSparse(
            psi,
            ham,
            numerics.U1_symmetry,
            init_bond_dim=numerics.initial_bond_dimension,
            max_bond_dim=numerics.max_bond_dimensions[0],
            energy_degeneracy_threshold=energy_degeneracy_threshold,
            entanglement_degeneracy_threshold=entanglement_degeneracy_threshold,
        )
        if gss.init_u1_num != gss.u1_num:
            sz_sign = np.sign(gss.u1_num - gss.init_u1_num)
            sz_diff = abs(gss.u1_num - gss.init_u1_num) // 2
            gss.run(
                opt_structure=0,
                max_num_sweep=sz_diff,
                energy_convergence_threshold=0.0,
                entanglement_convergence_threshold=0.0,
                sz_sign=sz_sign,
            )
            print(f"Complete the warmup sweeps for M={gss.u1_num}/2.")
    else:
        gss = GroundStateSearch(
            psi,
            ham,
            init_bond_dim=numerics.initial_bond_dimension,
            max_bond_dim=numerics.initial_bond_dimension,
            energy_degeneracy_threshold=energy_degeneracy_threshold,
            entanglement_degeneracy_threshold=entanglement_degeneracy_threshold,
        )

    for i, (max_bond_dim, max_num_sweep) in enumerate(
        zip(numerics.max_bond_dimensions, numerics.max_num_sweeps)
    ):
        gss.max_bond_dim = max_bond_dim
        if i == 0:
            gss.run(
                opt_structure=opt_structure,
                energy_convergence_threshold=energy_convergence_threshold,
                entanglement_convergence_threshold=entanglement_convergence_threshold,
                max_num_sweep=max_num_sweep,
                temperature=temperature,
                tau=tau,
            )
            print("Calculating the expectation values for the initial structure")
            # re-run the first iteration to save the expectation values
            gss.run(
                opt_structure=0,
                max_num_sweep=1,
                eval_onesite_expval=save_onesite_expval,
                eval_twosite_expval=save_twosite_expval,
            )
        else:
            gss.run(
                opt_structure=0,
                energy_convergence_threshold=energy_convergence_threshold,
                entanglement_convergence_threshold=entanglement_convergence_threshold,
                max_num_sweep=max_num_sweep,
                eval_onesite_expval=save_onesite_expval,
                eval_twosite_expval=save_twosite_expval,
            )

        nodes_list = {}
        for edge_id in gss.energy.keys():
            tmp = []
            for node_id, edges in enumerate(psi.edges):
                node_id += config.system.N
                if edge_id in edges:
                    tmp.append(node_id)
            nodes_list[edge_id] = tmp

        for edge_id in psi.physical_edges:
            tmp = []
            tmp.append(edge_id)
            for node_id, edges in enumerate(psi.edges):
                node_id += config.system.N
                if edge_id in edges:
                    tmp.append(node_id)
            nodes_list[edge_id] = tmp
            gss.energy[edge_id] = 0.0
            gss.error[edge_id] = 0.0

        all_keys = set(nodes_list.keys())
        df = pd.DataFrame(
            [nodes_list[k] for k in all_keys],
            columns=["node1", "node2"],
            index=None,
        )
        df["energy"] = [gss.energy[k] for k in all_keys]
        df["entanglement"] = [gss.entanglement[k] for k in all_keys]
        df["error"] = [gss.error[k] for k in all_keys]

        path_ = path / f"run{i + 1}"
        os.makedirs(path_, exist_ok=True)
        df.to_csv(path_ / "basic.csv", header=True, index=None)
        np.savetxt(path_ / "graph.dat", gss.psi.edges, fmt="%d", delimiter=",")

        if save_onesite_expval:
            df = pd.DataFrame(psi.physical_edges, columns=["site"], index=None)
            sp = np.zeros(len(psi.physical_edges))
            sm = np.zeros(len(psi.physical_edges))
            if not u1_symmetry:
                sp = np.array(
                    [
                        gss.one_site_expval[edge_id]["S+"]
                        for edge_id in psi.physical_edges
                    ]
                )
                sm = np.array(
                    [
                        gss.one_site_expval[edge_id]["S-"]
                        for edge_id in psi.physical_edges
                    ]
                )
            df["Sx"] = np.real((sp + sm) / 2.0)
            df["Sy"] = np.real((sp - sm) / 2.0j)
            df["Sz"] = np.real(
                [gss.one_site_expval[edge_id]["Sz"] for edge_id in psi.physical_edges]
            )

            path_ = path / f"run{i + 1}"
            os.makedirs(path_, exist_ok=True)
            df.to_csv(path_ / "single_site.csv", header=True, index=None)

        if save_twosite_expval:
            pairs = [(i, j) for i, j in itertools.combinations(psi.physical_edges, 2)]
            df = pd.DataFrame(pairs, columns=["site1", "site2"], index=None)
            spp = np.zeros(len(pairs))
            smm = np.zeros(len(pairs))
            szp = np.zeros(len(pairs))
            spz = np.zeros(len(pairs))
            szm = np.zeros(len(pairs))
            smz = np.zeros(len(pairs))
            if not u1_symmetry:
                spp = np.array([gss.two_site_expval[pair]["S+S+"] for pair in pairs])
                smm = np.array([gss.two_site_expval[pair]["S-S-"] for pair in pairs])
                szp = np.array([gss.two_site_expval[pair]["SzS+"] for pair in pairs])
                spz = np.array([gss.two_site_expval[pair]["S+Sz"] for pair in pairs])
                szm = np.array([gss.two_site_expval[pair]["SzS-"] for pair in pairs])
                smz = np.array([gss.two_site_expval[pair]["S-Sz"] for pair in pairs])

            szz = np.array([gss.two_site_expval[pair]["SzSz"] for pair in pairs])
            spm = np.array([gss.two_site_expval[pair]["S+S-"] for pair in pairs])
            smp = np.array([gss.two_site_expval[pair]["S-S+"] for pair in pairs])

            df["SxSx"] = np.real((spp + spm + smp + smm) / 4.0)
            df["SySy"] = np.real((-spp + spm + smp - smm) / 4.0)
            df["SzSz"] = np.real(szz)

            df["SySz"] = np.real((spz - smz) / 2.0j)
            df["SzSy"] = np.real((szp - szm) / 2.0j)

            df["SzSx"] = np.real((szp + szm) / 2.0)
            df["SxSz"] = np.real((spz + smz) / 2.0)

            df["SxSy"] = np.real((spp - spm + smp - smm) / 4.0j)
            df["SySx"] = np.real((spp + spm - smp - smm) / 4.0j)

            path_ = path / f"run{i + 1}"
            os.makedirs(path_, exist_ok=True)
            df.to_csv(path_ / "two_site.csv", header=True, index=None)

    return 0



import tensornetwork as tn
import numpy as np

from ttnopt.src.TTN import TreeTensorNetwork
from ttnopt.src.TwoSiteUpdater import TwoSiteUpdater
from ttnopt.src.functionTTN import (
    get_renormalization_sequence,
)

tn.set_default_backend("numpy")


class DataEngine(TwoSiteUpdater):
    def __init__(
        self,
        psi: TreeTensorNetwork,
        max_bond_dim: int,
    ):
        """Initialize a PhysicsEngine object.

        Args:
            psi (TreeTensorNetwork): The quantum state.
            target (np.ndarray): Tensor of target_data.
            max_bond_dim (int): Maximum bond dimension.
        """
        super().__init__(psi)
        self.max_bond_dim = max_bond_dim
        self.environment_tensor = None
        self.environment_edges = None

    def get_fidelity(self, target):
        sequence = get_renormalization_sequence(
            self.psi.edges, self.psi.canonical_center_edge_id
        )
        target_node = tn.Node(target)
        target_node = target_node.copy(conjugate=True)
        dangling_edges_dict = {i: target_node[i] for i in self.psi.physical_edges}
        for tensor_id in sequence:
            iso = tn.Node(self.psi.tensors[tensor_id])
            for i, edge_id in enumerate(self.psi.edges[tensor_id][:2]):
                dangling_edges_dict[edge_id] ^ iso[i]
                # remove dangling edge
                dangling_edges_dict.pop(edge_id)
            out_edge_orders = list(dangling_edges_dict.values()) + [iso[2]]
            target_node = tn.contractors.auto(
                [target_node, iso], output_edge_order=out_edge_orders
            )
            dangling_edges_dict[self.psi.edges[tensor_id][2]] = target_node[-1]
        gauge = tn.Node(self.psi.gauge_tensor)
        target_node[0] ^ gauge[0]
        target_node[1] ^ gauge[1]
        inner_prod = tn.contractors.auto([target_node, gauge])
        return np.abs(inner_prod.tensor) ** 2

    def update_tensor(self, target, central_tensor_ids):
        environment, out_edge_orders = self.environment(target, central_tensor_ids)
        output_order = (
            self.psi.edges[central_tensor_ids[0]][:2]
            + self.psi.edges[central_tensor_ids[1]][:2]
        )
        environment.reorder_edges(
            [out_edge_orders[edge_id] for edge_id in output_order]
        )
        return environment

    def environment(self, target, tensor_ids):
        sequence = get_renormalization_sequence(
            self.psi.edges, self.psi.canonical_center_edge_id
        )
        environment = tn.Node(target)
        environment = environment.copy(conjugate=True)
        dangling_edges_dict = {i: environment[i] for i in self.psi.physical_edges}
        for tensor_id in sequence:
            if tensor_id not in tensor_ids:
                iso = tn.Node(self.psi.tensors[tensor_id])
                for i, edge_id in enumerate(self.psi.edges[tensor_id][:2]):
                    dangling_edges_dict[edge_id] ^ iso[i]
                    # remove dangling edge
                    dangling_edges_dict.pop(edge_id)
                out_edge_orders = list(dangling_edges_dict.values()) + [iso[2]]
                environment = tn.contractors.auto(
                    [environment, iso], output_edge_order=out_edge_orders
                )
                dangling_edges_dict[self.psi.edges[tensor_id][2]] = environment[-1]
        return environment, dangling_edges_dict
